"""
-------------------------------------------------------------------------------

Design details of the high-frequency poly-ema level filter as it relates
to the jump-volatility filter.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.analog import polyema_level as dsgn_a_l_pema


def get_valid_filter_orders() -> np.ndarray:
    """Returns valid filter orders for this filter design"""

    # start with the analog design space
    analog_candidate = dsgn_a_l_pema.get_valid_filter_orders(strict=True)

    # remove orders <= 2
    valid_orders = analog_candidate[np.where(analog_candidate >= 3)[0]]

    # return
    return valid_orders


def designs(filter_order: int) -> dict:
    """Design details for the jump-vol filter"""

    # designs
    design = {}

    design[3] = {
        "kij_zero_at_tau_min": 1.06966,
        "kij_zero_at_tau_inf": 0.883706,
    }

    design[4] = {
        "kij_zero_at_tau_min": 1.06752,
        "kij_zero_at_tau_inf": 0.981931,
    }

    design[5] = {
        "kij_zero_at_tau_min": 1.13212,
        "kij_zero_at_tau_inf": 1.07404,
    }

    design[6] = {
        "kij_zero_at_tau_min": 1.2103,
        "kij_zero_at_tau_inf": 1.16004,
    }

    design[7] = {
        "kij_zero_at_tau_min": 1.28814,
        "kij_zero_at_tau_inf": 1.24068,
    }

    design[8] = {
        "kij_zero_at_tau_min": 1.36295,
        "kij_zero_at_tau_inf": 1.31675,
    }

    # return
    return design[filter_order]



"""
-------------------------------------------------------------------------------

Analog Bessel filter design:

Captures the reference pole locations and partial-fraction expansion residues
for analog Bessel filters, and also captures pole groupings for multistage
design.

Wavenumbers are calculated numerically in Python in `calculate_wavenumber.py`.

-------------------------------------------------------------------------------
"""

import numpy as np
import sys

from irides.resources.containers import points
from irides.tools import design_tools
from irides.tools.design_tools import StageTypes


# noinspection SpellCheckingInspection
def design_id() -> str:
    """
    Since `design` modules are passed around as objects in this framework,
    a `design_id` gives a unique name.

    Returns
    -------
    str
        The id of this design.
    """

    return "bessel-level"


def design_type() -> design_tools.FilterDesignType:
    """This is a level filter"""

    return design_tools.FilterDesignType.LEVEL


# noinspection SpellCheckingInspection,PyUnusedLocal
def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns valid filter orders for this filter design.

    Parameters
    ----------
    strict: bool
        When true, only orders where h(0)=0 are included.

    Returns
    -------
    np.ndarray
        Array of valid filter orders
    """

    return design_tools.get_valid_filter_orders(
        design_tools.FilterDesignType.LEVEL, strict
    )


"""
-------------------------------------------------------------------------------

Filter-specific definitions

-------------------------------------------------------------------------------
"""


# noinspection SpellCheckingInspection
def designs(filter_order: int) -> dict:
    r"""
    The mth order Bessel filter (Laplace) transfer function is defined by

                   a_{0,m}
        Hm(s) = -------------- ,
                 theta^(m)(s)

    where theta^(m)(s) is an mth order reverse Bessel polynomial defined by

                                     (2m - k)!      s^k
        theta^(m)(s) = sum_{k=0}^m ------------- ---------
                                    k! (m - k)!   2^{m-k}

    and
                   (2m)!
        a_{0,m} = -------- .
                   2^m m!

    Numeric values of roots and residues are computed by Wolfram Alpha, located
    at https://www.wolframalpha.com/. An example root evaluation string is:

        roots s^2 + 3 s + 3.

    The concept of stages stems from the observation that the a transfer
    function, such as

                          15
        H(s) = -------------------------,
                s^3 + 6 s^2 + 15 s + 15

    can be rewritten as

                  pr^2 + pi^2         pa
        H(s) = ----------------- . --------.
                (s+pr)^2 + pi^2     pa + s

    The inverse Laplace transform yields the multistage construction:

        h(t) = L^(-1){ H_dosc(s) } * L^(-1){ H_ema(s) }

    where * is the convolution operator.

    The stage definitions below adhere to the multistage method
    of construction.

    An alternative construction, which is not recommendedjn practice but
    will help with aspects of design, is a full partial-fraction expansion
    of H(s). Forjnstance, as per above,

                  A1         A2         A3
        H(s) = -------- + -------- + --------,
                s + p1     s + p2     s + p3

    where {p_k} are the pole locations and {A_k} are the residues. A residue
    is naturally associated with a specific pole. See file bessel.nb for the
    calculation of pole and residue values.

    Parameters
    ----------
    filter_order: int
        The order of the filter (note: user is required to validate that
        the order is supported).

    Returns
    -------
    dict
        Design dictionary with keys 'stages', 'poles', 'residues'.
    """

    # designs
    design = {}

    r"""
    1st order transfer function:

                  1
        H(s) = -------
                s + 1

    stage design:

             -------
        ---->| ema |---->
             -------

    """

    design[1] = {
        "stages": [{"type": StageTypes.EMA.value, "indices": np.array([0])}],
        "poles": np.array([-1.0 + 0j]),
        "residues": np.array([1]),
        "poly_coeffs": np.array([1, 1]),
    }

    r"""
    2nd order transfer function:

                      3
        H(s) = ---------------
                s^2 + 3 s + 3

    stage design:

             ---------
        ---->| d-osc |---->
             ---------

    """

    design[2] = {
        "stages": [
            {"type": StageTypes.DOSC.value, "indices": np.array([1, 0])}
        ],
        "poles": np.array(
            [
                -1.5000000000000000000 - 0.8660254037844386468j,
                -1.5000000000000000000 + 0.8660254037844386468j,
            ]
        ),
        "residues": np.array(
            [0.0 + 1.7320508075688772935j, 0.0 - 1.7320508075688772935j]
        ),
        "poly_coeffs": np.array([1, 3, 3]),
    }

    r"""
    3rd order transfer function:

                         15
        H(s) = -------------------------
                s^3 + 6 s^2 + 15 s + 15

    stage design:

             ---------     -------
        ---->| d-osc |---->| ema |---->
             ---------     -------

    """

    design[3] = {
        "stages": [
            {"type": StageTypes.DOSC.value, "indices": np.array([2, 1])},
            {"type": StageTypes.EMA.value, "indices": np.array([0])},
        ],
        "poles": np.array(
            [
                -2.3221853546260855929,
                -1.8389073226869572035 - 1.7543809597837216610j,
                -1.8389073226869572035 + 1.7543809597837216610j,
            ]
        ),
        "residues": np.array(
            [
                4.5297921600971586592,
                -2.2648960800485793296 + 0.62390925642938012061j,
                -2.2648960800485793296 - 0.62390925642938012061j,
            ]
        ),
        "poly_coeffs": np.array([1, 6, 15, 15]),
    }

    r"""
    4th order transfer function:

                               105
        H(s) = -------------------------------------
                s^4 + 10 s^3 + 45 s^2 + 105 s + 105

    stage design:

             ---------     ---------
        ---->| d-osc |---->| d-osc |---->
             ---------     ---------

    """

    design[4] = {
        "stages": [
            {"type": StageTypes.DOSC.value, "indices": np.array([3, 2])},
            {"type": StageTypes.DOSC.value, "indices": np.array([1, 0])},
        ],
        "poles": np.array(
            [
                -2.8962106028203721684 - 0.8672341289345037518j,
                -2.8962106028203721684 + 0.8672341289345037518j,
                -2.1037893971796278316 - 2.6574180418567527169j,
                -2.1037893971796278316 + 2.6574180418567527169j,
            ]
        ),
        "residues": np.array(
            [
                1.6633914159371089178 + 8.3963003152960348792j,
                1.6633914159371089178 - 8.3963003152960348792j,
                -1.6633914159371089178 - 2.2440773205454007801j,
                -1.6633914159371089178 + 2.2440773205454007801j,
            ]
        ),
        "poly_coeffs": np.array([1, 10, 45, 105, 105]),
    }

    r"""
    5th order transfer function:

                                   945
        H(s) = ------------------------------------------------
                s^5 + 15 s^4 + 105 s^3 + 420 s^2 + 945 s + 945

    stage design:

             ---------     ---------     -------
        ---->| d-osc |---->| d-osc |---->| ema |---->
             ---------     ---------     -------

    """

    design[5] = {
        "stages": [
            {"type": StageTypes.DOSC.value, "indices": np.array([4, 3])},
            {"type": StageTypes.DOSC.value, "indices": np.array([2, 1])},
            {"type": StageTypes.EMA.value, "indices": np.array([0])},
        ],
        "poles": np.array(
            [
                -3.6467385953296432597,
                -3.3519563991535331430 - 1.7426614161831977227j,
                -3.3519563991535331430 + 1.7426614161831977227j,
                -2.3246743031816452271 - 3.5710229203379764004j,
                -2.3246743031816452271 + 3.5710229203379764004j,
            ]
        ),
        "residues": np.array(
            [
                20.863330821504189555,
                -11.935117225355992568 + 6.3430888965861925994j,
                -11.935117225355992568 - 6.3430888965861925994j,
                1.5034518146038977909 - 2.6668146309020763827j,
                1.5034518146038977909 + 2.6668146309020763827j,
            ]
        ),
        "poly_coeffs": np.array([1, 15, 105, 420, 945, 945]),
    }

    r"""
    6th order transfer function:

                                        10395
        H(s) = ----------------------------------------------------------------
                s^6 + 21 s^5 + 210 s^4 + 1260 s^3 + 4725 s^2 + 10395 s + 10395

    stage design:

             ---------     ---------     ---------
        ---->| d-osc |---->| d-osc |---->| d-osc |---->
             ---------     ---------     ---------

    """

    design[6] = {
        "stages": [
            {"type": StageTypes.DOSC.value, "indices": np.array([5, 4])},
            {"type": StageTypes.DOSC.value, "indices": np.array([3, 2])},
            {"type": StageTypes.DOSC.value, "indices": np.array([1, 0])},
        ],
        "poles": np.array(
            [
                -4.2483593958633639449 - 0.8675096732313656064j,
                -4.2483593958633639449 + 0.8675096732313656064j,
                -3.7357083563258146679 - 2.6262723114471256405j,
                -3.7357083563258146679 + 2.6262723114471256405j,
                -2.5159322478108213871 - 4.4926729536539425359j,
                -2.5159322478108213871 + 4.4926729536539425359j,
            ]
        ),
        "residues": np.array(
            [
                10.959228792516907373 + 39.425157113160190552j,
                10.959228792516907373 - 39.425157113160190552j,
                -14.126767991748670862 - 12.701164165560234053j,
                -14.126767991748670862 + 12.701164165560234053j,
                3.1675391992317634894 + 0.20245898404171038568j,
                3.1675391992317634894 - 0.20245898404171038568j,
            ]
        ),
        "poly_coeffs": np.array([1, 21, 210, 1260, 4725, 10395, 10395]),
    }

    r"""
    7th order transfer function:

                                        135135
        H(s) = -----------------------------------------------------------------
                s^7 + 28 s^6 + 378 s^5 + 3150 s^4 + 17325 s^3 
                                                + 62370 s^2 + 135135 s + 135135

    stage design:

             ---------     ---------     ---------     -------
        ---->| d-osc |---->| d-osc |---->| d-osc |---->| ema |---->
             ---------     ---------     ---------     -------

    """

    design[7] = {
        "stages": [
            {"type": StageTypes.DOSC.value, "indices": np.array([6, 5])},
            {"type": StageTypes.DOSC.value, "indices": np.array([4, 3])},
            {"type": StageTypes.DOSC.value, "indices": np.array([2, 1])},
            {"type": StageTypes.EMA.value, "indices": np.array([0])},
        ],
        "poles": np.array(
            [
                -4.9717868585279356779,
                -4.7582905281546289452 - 1.7392860611305365429j,
                -4.7582905281546289452 + 1.7392860611305365429j,
                -4.0701391636381374717 - 3.5171740477097531658j,
                -4.0701391636381374717 + 3.5171740477097531658j,
                -2.6856768789432657441 - 5.4206941307167488958j,
                -2.6856768789432657441 + 5.4206941307167488958j,
            ]
        ),
        "residues": np.array(
            [
                96.448703466230703258,
                -57.792465959616429334 + 38.884808124790946028j,
                -57.792465959616429334 - 38.884808124790946028j,
                8.3218236759203368582 - 23.140605908290046126j,
                8.3218236759203368582 + 23.140605908290046126j,
                1.2462905505807408470 + 2.9043702547805305311j,
                1.2462905505807408470 - 2.9043702547805305311j,
            ]
        ),
        "poly_coeffs": np.array(
            [1, 28, 378, 3150, 17325, 62370, 135135, 135135]
        ),
    }

    r"""
    8th order transfer function:

                                        2027025
        H(s) = -----------------------------------------------------------------
                s^8 + 36 s^7 + 630 s^6 + 6930 s^5 + 51975 s^4 + 270270 s^3 
                                              + 945945 s^2 + 2027025 s + 2027025

    stage design:

             ---------     ---------     ---------     ---------
        ---->| d-osc |---->| d-osc |---->| d-osc |---->| d-osc |---->
             ---------     ---------     ---------     ---------


    """

    design[8] = {
        "stages": [
            {"type": StageTypes.DOSC.value, "indices": np.array([7, 6])},
            {"type": StageTypes.DOSC.value, "indices": np.array([5, 4])},
            {"type": StageTypes.DOSC.value, "indices": np.array([3, 2])},
            {"type": StageTypes.DOSC.value, "indices": np.array([1, 0])},
        ],
        "poles": np.array(
            [
                -5.5878860432630851990 - 0.8676144453527864598j,
                -5.5878860432630851990 + 0.8676144453527864598j,
                -5.2048407906368819183 - 2.6161751526425274287j,
                -5.2048407906368819183 + 2.6161751526425274287j,
                -4.3682892172024024070 - 4.4144425004715390836j,
                -4.3682892172024024070 + 4.4144425004715390836j,
                -2.8389839488976304757 - 6.3539112986048768221j,
                -2.8389839488976304757 + 6.3539112986048768221j,
            ]
        ),
        "residues": np.array(
            [
                59.136190765097912790 + 183.86040218976815559j,
                59.136190765097912790 - 183.86040218976815559j,
                -87.076721410343559811 - 62.455631732253360265j,
                -87.076721410343559811 + 62.455631732253360265j,
                29.865102171946156214 - 2.0276026150302011702j,
                29.865102171946156214 + 2.0276026150302011702j,
                -1.9245715267005091935 + 2.3681579888539021834j,
                -1.9245715267005091935 - 2.3681579888539021834j,
            ]
        ),
        "poly_coeffs": np.array(
            [1, 36, 630, 6930, 51975, 270270, 945945, 2027025, 2027025]
        ),
    }

    # check and return
    return design[filter_order]


"""
-------------------------------------------------------------------------------

Standard features

-------------------------------------------------------------------------------
"""


# noinspection SpellCheckingInspection
def unit_peak_time(filter_order: int) -> float:
    """
    The timepoint where a unit-scale impulse response h(t)
    reaches its first peak. See bessel_inline_slope.nb for calculation.

    Parameters
    ----------
    filter_order: int
        The order of the filter (note: user is required to validate that
        the order is supported).

    Returns
    -------
    float
        The timepoint where h(t) peaks, for unit delay.
    """

    # values calculated from Mathematica
    unit_peak_times = np.array(
        [
            0.000000000000000,  # m = 1
            0.604599788078073,  # m = 2
            0.813499894915707,  # m = 3
            0.900548102972202,  # m = 4
            0.942571467655463,  # m = 5
            0.964984859216068,  # m = 6
            0.977802459643315,  # m = 7
            0.985513959386379,  # m = 8
        ]
    )

    # return
    return unit_peak_times[filter_order - 1]


# noinspection SpellCheckingInspection
def unit_peak_value(filter_order: int) -> float:
    """
    The value of h(t_peak), the first and principal peak, for a unit-scale
    impulse response. See bessel_inline_slope.nb for calculation.

    Parameters
    ----------
    filter_order: int
        The order of the filter (note: user is required to validate that
        the order is supported).

    Returns
    -------
    float
        The value of h(t_peak), for unit delay.
    """

    # values calculated from Mathematica
    unit_peak_values = np.array(
        [
            1.000000000000000,  # m = 1
            0.699357279553692,  # m = 2
            0.816382165302566,  # m = 3
            0.952223206957527,  # m = 4
            1.085517577490184,  # m = 5
            1.212016261684917,  # m = 6
            1.331053039720972,  # m = 7
            1.442974600373275,  # m = 8
        ]
    )

    # return
    return unit_peak_values[filter_order - 1]


# noinspection SpellCheckingInspection,PyUnusedLocal
def peak_value_coordinate(tau: float, filter_order: int, dt=0.0):
    """
    The impulse response of a low-pass level Bessel filter
    has a principal mode for t > 0, and then rings slightly in the
    tail. The peak time /t*/ and peak value /h(t*)/, which look
    roughly like this,

                  ....x...  <---- h(t*)
              ....        ....
           ...                ..
         ..                     ...
        .                          ....
    ----|-------------|-------------------------> t
        0
                     t*

    are calculated numerically. See bessel_inline_slope.nb for the
    calculations. Bessel filters differ from polyemas in that polyemas
    have analytic expressions for any filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter (note: user is required to validate that
        the order is supported).
    dt: float
        Temporal discretization interval.

    Returns
    -------
    ContinuousTimePoint
        Object with .time and .value fields.
    """

    # aliases
    m = filter_order

    # put together into a point coordinate
    peak_coord = points.ContinuousTimePoint(
        unit_peak_time(m) * tau, unit_peak_value(m) / tau
    )

    # return
    return peak_coord


# noinspection SpellCheckingInspection,PyUnusedLocal
def moment_value_generator(moment: int, tau: float = 1.0, dt: float = 0.0):
    """
    Returns an anonymous function with the single argument (order: int),
    where order is a `filter_order`. The returned function captures the `tau`
    and `dt` arguments in this calling function.

    Note that the generated functions are not purely continuous time.
    Instead, they bridge continuous and discrete time by applying order-dt
    corrections to the moments. This is required because the CT series are
    in fact implemented in DT. Use dt=0.0 to recover pure CT moments.

    The empirically determined polynomial functions in filter-order were
    estimated by calculating the raw M* moments, comparing them with the
    theoretic moments, and fitting a polynomial through the difference.
    Specifically,

        x = (order + 1) / 2
        corrections(x) = (raw-moment(x) - theo-moment(x)) / (dt / 2)
        poly(x) = fit(x, corrections(x), poly-order).

    Note that only odd filter orders require a dt-order correction; even
    orders require no correction. This is because odd orders blend in an
    ema stage, and leading edge of the ema induces a dt-order correction.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # generator defs with dt capture
    # noinspection PyUnusedLocal
    def m0_gen(order: int) -> float:
        """
        A dt correction is required for odd orders, but not even orders.
        The dt correction was determined empirically and is expressed as
        a linear function of filter order.
        """
        x = (order + 1.0) / 2.0
        dt_scale_correction = (-0.322 + x * 1.3224) if (order % 2 == 1) else 0.0

        return 1.0 + ((dt / tau) / 2.0) * dt_scale_correction

    # noinspection PyUnusedLocal
    def m1_gen(order: int) -> float:
        """
        Like M0, there is a dt-order correction to M1 for odd orders.
        The correction, while not analytic, was determined empirically
        and fit to a linear function of filter order.
        """
        x = (order + 1.0) / 2.0
        dt_scale_correction = (-1.322 + x * 1.3224) if (order % 2 == 1) else 0.0

        return tau + (dt / 2.0) * dt_scale_correction

    # noinspection PyUnusedLocal
    def m2_gen(order: int) -> float:
        """
        M2 is best fit with a cubic expression in filter order `order`.
        """
        x = (order + 1.0) / 2.0
        dt_scale_correction = (
            (0.1280003 + x * (-0.732667 + x * (0.678 - x * 0.0733333)))
            if (order % 2 == 1)
            else 0.0
        )

        return (
            2.0 * order / (2.0 * order - 1.0) * np.power(tau, 2)
            + (dt / 2.0) * dt_scale_correction
        )

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection
def autocorrelation_peak_and_stride_values(
    tau: float, filter_order: int
) -> dict:
    """
    System autocorrelation peak and stride values. Peak value is kh(0),
    and the stride comes from kh(xi_5pct) / kh(0) = 5%.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter (note: user is required to validate that
        the order is supported).

    Returns
    -------
    dict
        Peak value (scaled by 1/tau) and 5% pct stride (scaled by tau)
    """

    # values from Mathematica
    unit_sacf_values = {
        1: {"kh0": 0.5000, "xi_5pct": 2.9957},
        2: {"kh0": 0.4994, "xi_5pct": 2.1882},
        3: {"kh0": 0.5999, "xi_5pct": 1.6556},
        4: {"kh0": 0.7034, "xi_5pct": 1.3604},
        5: {"kh0": 0.8027, "xi_5pct": 1.1733},
        6: {"kh0": 0.8949, "xi_5pct": 1.0473},
        7: {"kh0": 0.9779, "xi_5pct": 0.9594},
        8: {"kh0": 1.0563, "xi_5pct": 0.8909},
    }

    # fetch values for order and scale
    sacf_values = unit_sacf_values[filter_order]
    sacf_values["kh0"] /= tau
    sacf_values["xi_5pct"] *= tau

    # return
    return sacf_values


# noinspection SpellCheckingInspection,PyPep8Naming
def scale_correlation_stride_and_residual_values(
    tau: float, filter_order: int
) -> dict:
    """Scale decorrelation strides and residual correlations.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Scale decorrelation (scales as `tau`) and residual correlation
        (independent of `tau`).
    """

    # values from Mathematica
    unit_sscf_values = {
        1: {"sti_5pct": 1598.00, "residual": 0.05},
        2: {"sti_5pct": 22.9564, "residual": 0.05},
        3: {"sti_5pct": 8.64193, "residual": 0.05},
        4: {"sti_5pct": 5.46444, "residual": 0.05},
        5: {"sti_5pct": 4.17151, "residual": 0.05},
        6: {"sti_5pct": 3.48665, "residual": 0.05},
        7: {"sti_5pct": 3.06588, "residual": 0.05},
        8: {"sti_5pct": 2.78165, "residual": 0.05},
    }

    # fetch values for order and scale
    sscf_values = unit_sscf_values[filter_order]
    sscf_values["sti_5pct"] *= tau

    # return
    return sscf_values


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def full_width_generator(tau: float = 1.0, dt: float = 0.0):
    """
    Returns anonymous function for the full-width. `tau` and `dt` are
    captured here, and the anonymous function has one argument, (order: int),
    the filter order.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # noinspection PyUnusedLocal
    def fw_gen(order: int) -> float:
        """
        A quadratic function of filter order is sufficient to capture
        the order-based dt-scale correction. Only odd orders, where an ema
        stage exists, requires a correction. The functional form was determined
        empirically.
        """
        x = (order + 1.0) / 2.0
        dt_scale_correction = (
            (2.836 + x * (-2.16 - x * 0.676)) if (order % 2 == 1) else 0.0
        )

        return (
            2.0 * np.sqrt(1.0 / (2.0 * order - 1.0))
            + (dt / 2.0) * dt_scale_correction
        )

    return fw_gen


# noinspection PyUnusedLocal
def gain_at_dc(tau: float) -> float:
    """Filter gain at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return 1.0


# noinspection PyUnusedLocal
def phase_at_dc(tau: float) -> float:
    """Filter phase at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return 0.0


def group_delay_at_dc(tau: float) -> float:
    """Filter group delay at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return tau


# noinspection SpellCheckingInspection
def cutoff_frequency(tau: float, filter_order: int) -> float:
    """
    At w_cutoff,

        |H(w_cutoff; tau, m)|^2 = 1/2.

    The solution is not analytic so the values for order 1-8 are reported here
    for tau = 1. The frequencies scale as 1 / tau.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter  (note: user is required to validate that
        the order is supported).

    Returns
    -------
    float
        Radian frequency at cutoff.
    """

    # cutoff freqs (from Mathematica bessel_level.nb, tau=1)
    cutoff_frequencies = {
        1: 1.00,
        2: 1.36,
        3: 1.76,
        4: 2.11,
        5: 2.43,
        6: 2.70,
        7: 2.95,
        8: 3.18,
    }

    # check and return
    return cutoff_frequencies[filter_order] / tau


# noinspection PyUnusedLocal
def gain_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The cutoff gain is always sqrt(1/2). The purpose of this function is
    to adhere to the phase- and group-delay-cutoff api so that any of the
    gain/phase/group-delay values @ cutoff can be determined uniformly.
    """

    return np.sqrt(1.0 / 2.0)


# noinspection PyUnusedLocal
def phase_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The phase at cutoff frequency `wc`. Phase at cutoff is independent
    of temporal scale `tau`.

    Parameters
    ----------
    tau: float
        Unused but present for a consistent api
    filter_order: int
        Order of the filter  (note: user is required to validate that
        the order is supported).

    Returns
    -------
    float
        Phase at cutoff frequency, tau = 1.
    """

    # phase at wc (from Mathematica bessel_level.nb, tau=1)
    phases_at_cutoff = {
        1: -0.785,
        2: -1.297,
        3: -1.736,
        4: -2.109,
        5: -2.426,
        6: -2.703,
        7: -2.952,
        8: -3.180,
    }

    # check and return
    return phases_at_cutoff[filter_order]


# noinspection SpellCheckingInspection
def group_delay_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The group delay at cutoff frequency `wc`, and the group delay scales with
    the `tau`. The unit-scale group-delay values are evaluated by
    Mathematica.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter (note: user is required to validate that
        the order is supported).

    Returns
    -------
    float
        Group delay at cutoff.
    """

    # cutoff freqs (from Mathematica bessel_level.nb, tau=1)
    group_delays_at_cutoff = {
        1: 0.500,
        2: 0.809,
        3: 0.935,
        4: 0.982,
        5: 0.996,
        6: 0.999,
        7: 1.000,
        8: 1.000,
    }

    # check and return
    return group_delays_at_cutoff[filter_order] * tau


# noinspection SpellCheckingInspection
def wireframe(tau: float, filter_order: int):
    """Wireframe timepoint for this level filter.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter (unused)

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    return design_tools.generate_level_wireframe_from_spectra(
        sys.modules[__name__], tau, filter_order
    )


# noinspection SpellCheckingInspection
def wavenumber(filter_order: int) -> float:
    """Report the wavenumber for this filter as a function of order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The wavenumber.
    """

    wavenumbers = {
        1: 0.64778095,
        2: 0.75316630,
        3: 0.77811737,
        4: 0.78250289,
        5: 0.76528238,
        6: 0.74164804,
        7: 0.71391692,
        8: 0.68803663,
    }

    return wavenumbers[filter_order]


# noinspection SpellCheckingInspection
def uncertainty_product(filter_order: int) -> dict:
    """Dict of TW, BW, and UCP for this filter order.

    Parameters
    ----------
    filter_order: int
        The order of the filter

    Returns
    -------
    dict
        TW is temporal width, BW is spectral width, and UCP is the
        uncertainty product.
    """

    uncertainty_products = {
        1: {"TW": 0.5, "BW": np.inf, "UCP": np.inf},
        2: {"TW": 0.440959, "BW": 1.73205, "UCP": 0.763763},
        3: {"TW": 0.360416, "BW": 1.58114, "UCP": 0.569868},
        4: {"TW": 0.304435, "BW": 1.74456, "UCP": 0.531105},
        5: {"TW": 0.265071, "BW": 1.95133, "UCP": 0.517241},
        6: {"TW": 0.23622, "BW": 2.16293, "UCP": 0.510926},
        7: {"TW": 0.214274, "BW": 2.36861, "UCP": 0.507532},
        8: {"TW": 0.197061, "BW": 2.56501, "UCP": 0.505463},
    }

    return uncertainty_products[filter_order]



"""
-------------------------------------------------------------------------------

Digital KT-Bessel inline-curvature filter design.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.analog import bessel_inline_curvature as dsgn_alg_c_bssl
from irides.design.digital import bessel_level as dsgn_dig_l_bssl
from irides.tools import design_tools


def design_id() -> str:
    """Design id of this filter."""

    return "digital-{0}".format(dsgn_alg_c_bssl.design_id())


def design_type() -> design_tools.FilterDesignType:
    """This is a curvature filter"""

    return dsgn_alg_c_bssl.design_type()


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns valid filter orders for this filter design."""

    return dsgn_alg_c_bssl.get_valid_filter_orders(strict)


def minimum_mu_value(filter_order: int) -> float:
    """Returns the minimum mu value for a given filter order"""

    return dsgn_dig_l_bssl.minimum_mu_value(filter_order)


def initial_hn_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns analytically calculated values of h[0], h[1], and h[2]

    Since this is an inline-curvature filter, h[k] values can be derived from
    the associated level-filter h[k] values.
    """

    # retrieve level-filter hk values
    hk_level = dsgn_dig_l_bssl.initial_hn_values(mu, filter_order)

    # compute hk_slope values
    h0 = hk_level[0]
    h1 = hk_level[1] - 2 * hk_level[0]
    h2 = hk_level[2] - 2 * hk_level[1] + hk_level[0]

    # pack and return
    return np.array([h0, h1, h2])


# noinspection PyPep8Naming
def moment_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns analytically calculated moment values (M0, M1, M2)

    In general, for an inline-slope filter,
        M0 = 0
        M1 = 0
        M2 = 2H(1).
    Rather than calculate H^k(1) terms again, they can be backed out from
    the level-filter calculations.
    """

    # retrieve level-filter Mk values
    Mk_level = dsgn_dig_l_bssl.moment_values(mu, filter_order)

    # convert Mk_level values to Mk_slope values
    M0 = 0
    M1 = 0
    M2 = 2.0 * Mk_level[0]

    # pack and return
    return np.array([M0, M1, M2])



"""
-------------------------------------------------------------------------------

Analog Bessel filter signature:

Implements the temporal and spectral signatures of a continuous-time
Bessel filter.

-------------------------------------------------------------------------------
"""

from typing import Callable
import numpy as np
from numpy.polynomial import polynomial
import sys

from irides.design.analog import bessel_level as design
from irides.design.analog import polyema_level as pema_design
from irides.filter_signatures.analog import polyema_level as pema_f_sig
from irides.filter_signatures.analog import damped_oscillator as dosc_f_sig
from irides.tools.design_tools import StageTypes
from irides.tools import impulse_response_builder_tools
from irides.tools import impulse_response_tools
from irides.tools import transfer_function_tools


# noinspection SpellCheckingInspection
def generate_impulse_response(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    """
    There are at least three ways to compute the Bessel-filter impulse response:

        1) simple partial-fraction expansion,
        2) multi-stage partial-fraction expansion, and
        3) multi-stage cascade.

    This function implements the multi-stage cascade form because it is the
    most numerically stable of the methods, and it is revealing to see how the
    final impulse-response is built from each additional stage.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order .

    Returns
    -------
    np.ndarray
        A time series array (n, 2 + n_stgs) having columns
        [t_axis, h_bessel, h_stg1, h_stg1 * h_stg2, ...].
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # fetch the design config for this filter order
    design_config = design.designs(filter_order)

    # aliases
    stages = design_config["stages"]
    poles = design_config["poles"]

    # calc the number of stages
    n_stgs = len(design_config["stages"])

    # make time series array, an (n, 2 + n_stgs)) matrix
    # the cols are [t, h(t), h_stg1(t), h_stg1(t) * h_stg2(t), ...]
    # where the cascaded convolution
    # make time series template
    (
        time_series,
        i_t_0plus,
    ) = impulse_response_builder_tools.make_impulse_response_template(
        t_start, t_end, dt, 2 + n_stgs
    )

    # aliases
    t_axis = time_series[:, 0]
    t_0plus = t_axis[i_t_0plus:]
    h = time_series[:, 1]

    # set effective t-start and -end points
    t_start_eff = min(t_0plus)
    t_end_eff = max(t_0plus)

    # create ht-generating fcxns

    # noinspection SpellCheckingInspection
    def ht_ema_gen(reference_poles: np.ndarray) -> np.ndarray:
        """Generator of h(t) for the ema. Captures scope-fixed arg values."""

        scaled_reference_tau = (
            tau
            * pema_design.convert_reference_poles_to_temporal_scales(
                reference_poles
            )[0]
        )
        return pema_f_sig.generate_impulse_response(
            t_start_eff, t_end_eff, dt, scaled_reference_tau, 1
        )[:, 1]

    # noinspection SpellCheckingInspection
    def ht_dosc_gen(reference_poles: np.ndarray) -> np.ndarray:
        """Generator of h(t) for the dosc. Captures scope-fixed arg values."""

        return dosc_f_sig.generate_impulse_response(
            t_start_eff, t_end_eff, dt, reference_poles[1], tau
        )[:, 1]

    # pack into a dict
    ht_gen_dict = {
        StageTypes.EMA.value: ht_ema_gen,
        StageTypes.DOSC.value: ht_dosc_gen,
    }

    # iter across stages
    ht_cand = np.array([1.0 / dt])  # starting point for cascade
    for i, stage in enumerate(stages):

        # gen h(t) for this stage (dosc or ema)
        ht_stage = ht_gen_dict[stage["type"]](poles[stage["indices"]])

        # convolve in new h(t) with running candidate
        ht_cand = np.convolve(ht_stage, ht_cand)[: t_0plus.shape[0]] * dt

        # persist running conv in an extended column
        time_series[i_t_0plus:, 2 + i] = ht_cand

    # persist the final cascade h(t) result
    h[i_t_0plus:] = ht_cand

    # return
    return time_series


# noinspection SpellCheckingInspection
def generate_sacf_correlogram(
    xi_start: float,
    xi_end: float,
    dxi: float,
    tau: float,
    filter_order: int,
    normalize=False,
) -> np.ndarray:

    # fetch impulse response
    ir = generate_impulse_response(0.0, xi_end, dxi, tau, filter_order)

    # calc sacf
    sacf = impulse_response_tools.calculate_auto_correlation_function(ir)

    # truncate at xi_start
    sacf = sacf[xi_start - dxi / 2.0 <= sacf[:, 0], :]

    # normalize by numeric peak value if requested
    # note that the numeric value will diverge from the theo value,
    # and the diveragence is a function of dxi, so it's better in this case
    # to normalize by the numerically calculated kh0 value.
    # note also that a xi = 0 value is expected, else the results will be
    # in error.
    if normalize:
        i_zero = np.where(0.0 - dxi / 2 < sacf[:, 0])[0][0]
        sacf[:, 1] /= sacf[i_zero, 1]

    return sacf


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_transfer_function(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    """
    Computes the transfer function H_bessel(f) from an array of frequencies f.
    The Bessel-filter transfer function is determined by the Bessel polynomial
    values, and these values depend on the filter order.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        The transfer function with cols [f_axis, H_bessel(f)].
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # make transfer-function panel
    npts = np.arange(f_start, f_end, df).shape[0]
    H_bessel = np.ndarray([npts, 2], dtype=complex)

    # setup f_axis
    H_bessel[:, 0] = np.arange(f_start, f_end, df)
    f_axis = H_bessel[:, 0]

    # make f_axis and jwtau vector
    jwtau = (1j * 2.0 * np.pi * tau) * f_axis

    # fetch Bessel denominator coeffs, set numerator coeff
    denom_coeffs = design.designs(filter_order)["poly_coeffs"]
    num_coeff = denom_coeffs[-1]

    # calculate the transfer function
    H_bessel[:, 1] = num_coeff / polynomial.polyval(jwtau, denom_coeffs[::-1])

    # return
    return H_bessel


# noinspection SpellCheckingInspection,PyPep8Naming
def make_callable_transfer_function(
    filter_order: int,
) -> Callable[[float, float], complex]:
    """Returns a callable function with signature (f, tau) -> complex.

    This transfer function is restricted to the jw axis:

        H(s) -> H(j w).

    Parameters
    ----------
    filter_order: int
        The filter order, which is captured by the definition herein.

    Returns
    -------
    Callable[[float, float], complex]
        A callable function (f: float, tau: float) -> complex
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # create capture parameters
    poly_coefs = design.designs(filter_order)["poly_coeffs"]
    num_coef = poly_coefs[-1]
    denom_coefs = poly_coefs[::-1]

    # noinspection SpellCheckingInspection,PyPep8Naming
    def H(f: float, tau: float = 1.0) -> complex:
        """Compute complex xfer fcxn H at freq `f` with optional scale tau.

        Note that `filter_order` is captured, leaving only (f, tau) as free
        variables.

        Parameters
        ----------
        f: float
            Start cyclic frequency.
        tau: float
            Temporal scaling, which ='s the first moment of the final filter.

        Returns
        -------
        complex
            H(f, tau) value.
        """

        jw = 1j * 2.0 * np.pi * f

        return num_coef / polynomial.polyval(jw * tau, denom_coefs)

    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_sscf_correlogram(
    sti_start: float, sti_end: float, dsti: float, filter_order: int
) -> np.ndarray:
    """
    Computes the normalized system scale-correlation correlogram from the
    filter's transfer function.

    No sti-normalization correction is necessary for level filters.

    Parameters
    ----------
    sti_start: float
        Start scale, |sti_start| <= |sti_end|
    sti_end: float
        End scale
    dsti: float
        Scale increment
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        System scale-correlation normalized (n, 2)
        np.array w/ cols [sti, sscf-value].
    """

    # correction function
    sti_norm_correction_fcxn = lambda sti: 1.0

    # return sscf-norm np-array
    return transfer_function_tools.calculate_scale_correlation_function(
        sti_start,
        sti_end,
        dsti,
        filter_order,
        sys.modules[__name__],
        sti_norm_correction_fcxn,
    )


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_phase_spectrum(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    """
    The phase of the Bessel-filter transfer function is the sum of phases from
    each stage.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        The transfer function with cols [f_axis, angle_H_bessel(f)].
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # make transfer-function panel
    npts = np.arange(f_start, f_end, df).shape[0]
    angle_H_bessel = np.zeros((npts, 2))

    # setup f_axis
    angle_H_bessel[:, 0] = np.arange(f_start, f_end, df)

    # fetch the design
    this_design = design.designs(filter_order)

    stages = this_design["stages"]
    poles = this_design["poles"]

    # iter over stages
    for stage in stages:

        stage_poles = poles[stage["indices"]]

        if stage["type"] == StageTypes.EMA.value:

            stage_order = 1
            stage_pole_taus = pema_design.convert_reference_poles_to_temporal_scales(
                stage_poles
            )
            stage_tau = tau * stage_pole_taus[0]

            angle_H_bessel[:, 1] += pema_f_sig.generate_phase_spectrum(
                f_start, f_end, df, stage_tau, stage_order
            )[:, 1]

        if stage["type"] == StageTypes.DOSC.value:

            reference_pole = stage_poles[0]

            angle_H_bessel[:, 1] += dosc_f_sig.generate_phase_spectrum(
                f_start, f_end, df, reference_pole, tau
            )[:, 1]

    # return
    return angle_H_bessel


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_spectra(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    """Generate gain, phase and group-delay spectra.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        Spectra panel [n_pts, 4] with cols [f, g(f), phase(f), grpdly(f)].
    """

    # fetch the transfer function for gain calculation
    H: np.ndarray = generate_transfer_function(
        f_start, f_end, df, tau, filter_order
    )

    # fetch the phase spectrum for group-delay calculation
    phase_spectrum: np.ndarray = generate_phase_spectrum(
        f_start, f_end, df, tau, filter_order
    )

    # return spectra
    return transfer_function_tools.calculate_spectra(H, phase_spectrum)


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_wireframe(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    """Returns a wireframe equivalent for this filter.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    WireframeContinuousTime
        A wireframe container.
    """

    # fetch the wireframe, create a timeseries template
    wf = design.wireframe(tau, filter_order)
    ts = impulse_response_builder_tools.discretize_wireframe_onto_time_grid(
        t_start, t_end, dt, wf
    )

    return ts


# noinspection SpellCheckingInspection,PyUnusedLocal
def generate_poles_and_zeros(
    tau: float, filter_order: int, f_start=0.0, f_end=0.0
) -> dict:
    """
    The Bessel-filter unit-scale poles are available in the design module, yet
    for consistency of the api, this function packs the poles for a particular
    order in a container. Continuous-time Bessel filters have no zeros.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.
    f_start: float
        Start cyclic frequency (not used in this signature).
    f_end: float
        End cyclic frequency (not used in this signature).

    Returns
    -------
    dict
        Dictionary with keys ('poles', 'zeros') and values as np.ndarray
        with dtype=complex.
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # fetch the design config for this filter order
    design_config = design.designs(filter_order)

    # scale poles by tau
    poles = design_config["poles"] / tau

    # fill in poles_zeros dict
    poles_zeros = {"poles": poles, "zeros": np.array([], dtype=complex)}

    # return
    return poles_zeros



"""
-------------------------------------------------------------------------------

Tools to build impulse responses and autocorrelation functions.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.resources.core_enumerations import FilterDesignType
from irides.resources.containers.wireframes import WireframeContinuousTime


# noinspection SpellCheckingInspection
def make_impulse_response_template(
    t_start: float, t_end: float, dt: float, n_cols: int = 2
) -> tuple:
    """Make a common impulse-response template as an np.ndarray [n, 2].

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    n_cols: int
        Number of columns (default = 2)

    Returns
    -------
    tuple
        0: np.ndarray time series template [n, 2] panel
        1:  index into time_axis from which all times are >= 0.0
    """

    # make time series, an (n, 2) matrix
    time_series = np.zeros(
        (np.arange(t_start, t_end + dt, dt).shape[0], n_cols)
    )

    # make t_axis
    time_series[:, 0] = np.arange(t_start, t_end + dt, dt)

    # alias
    t_axis = time_series[:, 0]

    # index and alias t_axis for t_axis >= 0
    i_t_0plus = protected_index_search(t_axis, dt, 0.0)

    # return tuple
    return time_series, i_t_0plus


# noinspection SpellCheckingInspection
def compute_noncausal_symmetric_first_difference(
    time_series: np.ndarray, enforce_causality=True
) -> int:
    """Note that time_series is changed in place, the original being clobbered.

    Calculates the noncausal, symmetric first difference for the time series.

       h(t)
        ^                   .  .  .
        |             . x .
        |           x |
                  .   |
        - . . . |---------------------------------> t
                0     |
                      v
        dh(t) = (h(t+dt) - h(t-dt)) / (2 dt)

    Note that h(t=0) is not reliable because t may not be indexed below zero,
    so a subsequent call to design.impulse_response_to_value(..) may be
    necessary.

    Parameters
    ----------
    time_series: np.ndarray
        Input time series, [n, n_cols >= 2].
    enforce_causality: bool
        Sets h[t<0] = 0 when true.

    Returns
    -------
    int:
        i_t_0plus as int or None
    """

    def compute_difference(ts: np.ndarray, dt: float) -> np.ndarray:
        """Implementation of central calculation"""

        # copy required so that resultant values are purely based on original
        # data, not a mixture of original and resultant, which is the case
        # since, otherwise, `h` is a reference to `ts[:, 1]`.
        h = ts[:, 1].copy()

        # difference
        scale_adj: float = 2.0 * dt
        ts[1:-1, 1] = (h[2:] - h[:-2]) / scale_adj
        ts[0, 1] = h[1] / scale_adj

        # last point cannot be calculated from a partial difference
        # b/c the level of the partial difference and a complete difference
        # may be large. Here, the last point is a copy of the preceding point.
        # An alternative is to return a shortened array. Returning an np.nan
        # disrupts downstream summation calculations.
        ts[-1, 1] = ts[-2, 1].copy()

        return ts

    return compute_noncausal_symmetric_generic_difference(
        compute_difference, time_series, enforce_causality
    )


# noinspection SpellCheckingInspection
def compute_noncausal_symmetric_second_difference(
    time_series_orig: np.ndarray, enforce_causality=True
) -> int:
    """Note that time_series is changed in place, the original being clobbered.

    Calculates the noncausal, symmetric second difference for the time series.

       h(t)
        ^                   .  .  .
        |             x x .
        |           x |
                  .   |
        - . . . |---------------------------------> t
                0     |
                      v
        dh(t) = (h(t+dt) - 2 h(t) + h(t-dt)) / dt^2

    Parameters
    ----------
    time_series_orig: np.ndarray
        Input time series, [n, n_cols >= 2].
    enforce_causality: bool
        Sets h[t<0] = 0 when true.

    Returns
    -------
    int:
        i_t_0plus as int or None
    """

    def compute_difference(ts: np.ndarray, dt: float) -> np.ndarray:
        """Implementation of central calculation"""

        # copy required so that resultant values are purely based on original
        # data, not a mixture of original and resultant, which is the case
        # since, otherwise, `h` is a reference to `ts[:, 1]`.
        h = ts[:, 1].copy()

        # difference
        scale_adj = np.power(dt, 2)
        ts[1:-1, 1] = (h[2:] - 2.0 * h[1:-1] + h[:-2]) / scale_adj
        ts[0, 1] = (h[1] - 2.0 * h[0]) / scale_adj

        # last point cannot be calculated from a partial difference
        # b/c the level of the partial difference and a complete difference
        # may be large. Here, the last point is a copy of the preceding point.
        # An alternative is to return a shortened array. Returning an np.nan
        # disrupts downstream summation calculations.
        ts[-1, 1] = ts[-2, 1].copy()

        return ts

    return compute_noncausal_symmetric_generic_difference(
        compute_difference, time_series_orig, enforce_causality
    )


# noinspection SpellCheckingInspection
def compute_noncausal_symmetric_generic_difference(
    compute_difference, time_series: np.ndarray, enforce_causality=True
) -> int:
    """Computes nth difference, preserves causality (if req'd)"""

    # features along the time axis
    t_axis = time_series[:, 0]
    dt = np.min(np.diff(t_axis))
    i_t0_plus = protected_index_search(t_axis, dt, 0.0)

    # compute difference
    time_series = compute_difference(time_series, dt)

    # force set h(t<0) = 0
    if enforce_causality and i_t0_plus is not None:
        time_series[:i_t0_plus, 1] = 0.0

    # return
    return i_t0_plus


# noinspection SpellCheckingInspection
def protected_index_search(
    axis: np.ndarray, step: float, search_value: float
) -> int:
    """Returns index into axis at `search_value` or None.

    Parameters
    ----------
    axis: np.ndarray
        Axis to search.
    step: float
        Step size along axis.
    search_value

    Returns
    -------
    int:
        index value or None
    """

    # search
    candidate = np.where(search_value - step / 2.0 <= axis)[0]

    # protected return
    return candidate[0] if candidate.shape[0] > 0 else None


# noinspection SpellCheckingInspection
def discretize_wireframe_onto_time_grid(
    t_start: float, t_end: float, dt: float, wf: WireframeContinuousTime
) -> np.ndarray:
    """Embeds a continuous-time wireframe onto a discretized time grid.

    Parameters
    ----------
    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    wf: WireframeContinuousTime
        A populated wireframe object

    Returns
    -------
    np.ndarray
        [n, 2] time-series array
    """

    # make ts template, take alias of the time axis
    ts, _ = make_impulse_response_template(t_start, t_end, dt)
    t_axis = ts[:, 0]

    # for each delta function, place it on the time grid
    tp_grid_indices = np.array([], dtype=int)
    for i, tp in enumerate(wf.timepoints):

        # check that this timepoint is in the [t_start, t_end] range
        if tp < t_start or tp > t_end:
            print(
                "warning: wireframe timepoint not in [t_start, t_end] window."
            )
            return ts

        # compute L1 norm between all t_axis points and tp, take shortest norm
        i_tp = np.abs(tp - t_axis).argmin()
        tp_grid_indices = np.append(tp_grid_indices, i_tp)

        # set delta (pre gain adjust)
        ts[i_tp, 1] = wf.weights[i]

    # correct for gain given the discretization
    if wf.type == FilterDesignType.LEVEL:
        gain_adj = 1.0
        ts[tp_grid_indices[0], 1] *= gain_adj

    if wf.type == FilterDesignType.SLOPE:
        wf_interval = t_axis[tp_grid_indices[1]] - t_axis[tp_grid_indices[0]]
        gain_adj = 1.0 / wf_interval
        ts[tp_grid_indices[0], 1] = +1.0 * gain_adj
        ts[tp_grid_indices[1], 1] = -1.0 * gain_adj

    if wf.type == FilterDesignType.CURVE:
        wf_interval = t_axis[tp_grid_indices[2]] - t_axis[tp_grid_indices[0]]
        gain_adj = np.power(wf_interval / 2.0, -2)

        ts[tp_grid_indices[0], 1] = +1.0 * gain_adj
        ts[tp_grid_indices[1], 1] = -2.0 * gain_adj
        ts[tp_grid_indices[2], 1] = +1.0 * gain_adj

    # universal dt adjustment
    ts[:, 1] /= dt

    return ts



"""
-------------------------------------------------------------------------------

Tools to convert analog poles to digital, provide for pole-location scaling,
and offer return in either Cartesian or polar form.

-------------------------------------------------------------------------------
"""

import numpy as np
import scipy.optimize

from typing import Union


# noinspection SpellCheckingInspection,PyPep8Naming
def convert_analog_poles_to_digital_poles(
    tau: float, splane_poles: np.ndarray, T: float = 1
) -> np.ndarray:
    """Converts s-plane poles via z = e^(p T / tau) to digital poles.

    Parameters
    ----------
    tau: float
        Scale factor, generally tau > 0
    splane_poles: np.ndarray
        Poles in the s-plane
    T: float
        Sampling period, use default T=1 except for special cases.

    Returns
    -------
    np.ndarray
        Converted z-plane poles (Cartesian coords).
    """

    # convert via exponential map and return
    return np.exp(splane_poles * T / tau)


# noinspection SpellCheckingInspection,PyPep8Naming
def convert_digital_poles_to_analog_poles(
    zplane_poles: np.ndarray, T: float = 1
) -> np.ndarray:
    """Converts z-plane poles via s = log(p / T) to analog poles.

    Parameters
    ----------
    zplane_poles: np.ndarray
        Poles in the z-plane
    T: float
        Sampling period, use default T=1 except for special cases.

    Returns
    -------
    np.ndarray
        Converted z-plane poles (Cartesian coords).
    """

    # convert via logarithmic map and return
    return np.log(zplane_poles / T)


# noinspection SpellCheckingInspection
def first_moment_from_zplane_pole(
    zplane_poles: Union[complex, np.ndarray],
) -> Union[float, np.ndarray]:
    """Infer M1 values from z-plane pole coords |p|.

    Parameters
    ----------
    zplane_poles: np.ndarray
        Poles in the z-plane

    Returns
    -------
    np.ndarray
        inferred M1-d values
    """

    # compute and return
    p_abs = np.abs(zplane_poles)
    return p_abs / (1.0 - p_abs)


# noinspection SpellCheckingInspection,PyPep8Naming
def zplane_poles_from_analog_design(
    design_alg, mu: float, filter_order: int, T: float = 1
) -> np.ndarray:
    """Returns zplane poles provided a given design.

    There is a common pattern where
        splane-poles    <- design
        tau             <- calc_from(mu, splane-poles)
        zplane-poles    <- calc_from(splane-poles, tau),

    and so this function captures it.

    Parameters
    ----------
    design_alg:
        A design module that has designs() as a function.
    mu: float
        Digital scale parameter
    filter_order: int
        Filter order
    T: float
        Sampling period (defaults to 1)

    Returns
    -------
    np.ndarray
        Array of zplane poles
    """

    # run through the sequence
    splane_poles = design_alg.designs(filter_order)["poles"]
    tau = solve_for_tau_from_mu_and_splane_poles(mu, splane_poles, T)
    zplane_poles = convert_analog_poles_to_digital_poles(tau, splane_poles)

    # return
    return zplane_poles


# noinspection SpellCheckingInspection,PyPep8Naming
def solve_for_tau_to_place_splane_pole_onto_zplane_real_axis(
    splane_pole: complex, zplane_real_coordinate: float, T: float = 1
) -> float:
    """Solves for `tau` s.t. Re(e^(pa T / tau)) == rho.

    Solves f(tau) == 0

    where

        f(tau) = exp(par T / tau) cos(pai T / tau) - rho

    where rho = zplane_real_coordinate = real(pd). Invokes the `brentq` solver.

    `pa` is an s-plane pole,
    `pd` is the corresponding z-plane pole.

    Edge cases are treated analytically.

    Parameters
    ----------
    splane_pole: complex
        The s-plane pole to scale
    zplane_real_coordinate: float
        The value `rho` on the real z-plane axis, rho in [0, 1)
    T: float
        The sampling rate (default = 1)

    Returns
    -------
    float
        The scaling `tau` value that attains the goal, `tau` > 0.
    """

    # test that the zplane-coord is achievable
    if zplane_real_coordinate < 0 or zplane_real_coordinate >= 1:
        msg = (
            "The real-part of the z-plane coord must be on [0, 1), "
            "but was called with {0}".format(zplane_real_coordinate)
        )
        raise RuntimeError(msg)

    # pull of real and imaginary parts of the s-plane pole
    parT = T * np.real(splane_pole)
    paiT = T * np.abs(np.imag(splane_pole))

    # bounds on tau
    tau_upper = np.power(10, 9)
    tau_lower = paiT / (np.pi / 2.0)

    # analytic solutions:

    # imag(splane-pole) == 0 case (real-valued pa pole)
    if paiT <= 1e-12:
        soln = (
            parT / np.log(zplane_real_coordinate)
            if zplane_real_coordinate > 0
            else 0.0
        )
        return soln

    # rho == 0 case
    if np.abs(zplane_real_coordinate) <= 1e-12:
        return tau_lower

    # numeric solution:

    # the function for which f(tau) == 0 for optimal tau
    def f(tau: float) -> float:
        return np.exp(parT / tau) * np.cos(paiT / tau) - zplane_real_coordinate

    # call the solver
    tau_rho, r = scipy.optimize.brentq(
        f, tau_lower, tau_upper, full_output=True
    )

    return tau_rho


# noinspection PyPep8Naming
def compute_mu_from_tau_and_splane_poles(
    tau: Union[float, np.ndarray], splane_poles: np.ndarray, T: float = 1
) -> np.ndarray:
    """Compute `mu` from `tau` and analog poles.

    Parameters
    ----------
    tau: np.ndarray
        Temporal scales
    splane_poles: np.ndarray
        Analog poles in the s-plane
    T: float
        Sampling period (default = 1)

    Returns
    -------
    np.ndarray
        Value of `mu` to achieve target `tau`
    """

    # compute and return
    return np.sum(
        np.real(1.0 / (np.exp(-np.outer(splane_poles * T, 1.0 / tau)) - 1.0)),
        axis=0,
    )


# noinspection PyPep8Naming
def solve_for_tau_from_mu_and_splane_poles(
    mu: float, splane_poles: np.ndarray, T: float = 1
) -> float:
    """Computes `tau` such that M1(H(z)) = mu.

    This function computes an inflated `tau` value so that the target `mu`
    value is achieved.

    In general M1( H(z; mu) ) < M1( H(s; tau) ). The interface for digital
    filters uses scale-factor `mu` but the synthesis of these filters relies
    on the continuous-time design where the interface is in `tau`.

    Parameters
    ----------
    mu: float
        Temporal scale
    splane_poles: np.ndarray
        Analog poles in the s-plane
    T: float
        Sampling period (default = 1)

    Returns
    -------
    float
        Value of `tau` to achieve target `mu`
    """

    # set up
    filter_order = splane_poles.shape[0]

    # central function to drive to zero
    def f(tau: float) -> float:
        return compute_mu_from_tau_and_splane_poles(tau, splane_poles, T) - mu

    # setup bounds
    tau_lower = max(1.0, mu)
    tau_upper = filter_order + tau_lower

    # compute optimum tau
    tau_star, r = scipy.optimize.brentq(
        f, tau_lower, tau_upper, full_output=True
    )

    # return
    return tau_star


# noinspection SpellCheckingInspection,PyPep8Naming
def solve_for_tau_min_from_splane_poles(
    splane_poles: np.ndarray, T: float = 1
) -> dict:
    """Computes `tau_min` for a constellation of splane poles in a simple way.

    In this calculation, the outermost complex pole is placed in the zplane
    so that its projection onto the real axis lies as Re(pd) = 1/2. All poles
    inside of the outer pole will have real-axis projections that are larger.

    Parameters
    ----------
    splane_poles: np.ndarray
        Analog poles in the s-plane
    T: float
        Sampling period (default = 1)

    Returns
    -------
    dict
        Result dict, keys -- "tau_min" and "mu_min"
    """

    # identify pole with maximum imaginary part
    outer_splane_pole = splane_poles[np.argmax(np.imag(splane_poles))]

    # solve for tau value to place the real part of outer_splane_pole
    #         at Re(z) = 1/2.
    tau_min = solve_for_tau_to_place_splane_pole_onto_zplane_real_axis(
        outer_splane_pole, 0.5, T
    )

    # compute associated mu-min
    mu_min = compute_mu_from_tau_and_splane_poles(tau_min, splane_poles)[0]

    # return
    return {
        "tau_min": tau_min,
        "mu_min": mu_min,
    }


# noinspection SpellCheckingInspection,PyPep8Naming
def solve_for_min_tau_min_rho_tuple_from_splane_poles(
    splane_poles: np.ndarray, T: float = 1
) -> dict:
    """Solves for min(tau, a) tuple s.t. real-parts of z-plane poles are equal.

    The mapping of a constellation of s-plane poles, such as those of a
    classical, rational-function IIR filter, to the z-plane generally leads
    to a constellation with rightward or leftward concavity. This solver
    find min(tau) and min(a) such that the real-parts of the z-plane poles
    fall on a vertical line. This in turn produces the minimum scale-factor
    `tau` for which the discrete-time filter is not significantly aliased.

    From the s-plane constellation, the poles with the least and greatest
    imaginary parts are identified, and these are labeled p-inner and p-outer,
    respectively.

    Then, the iterative solution cycles through the sequence

        M_outer -> rho_outer --(p-outer)-->
                                tau
                                        --(p-inner)--> rho_inner -> M_inner.

    After a pass, M_outer' = M_inner, and the cycle repeats. Convergence is
    attained when |M_outer / M_inner - 1| < TOL. The sequence is initialized
    with a large M_outer value.

    Parameters
    ----------
    splane_poles: np.ndarray
        Constellation of s-plane poles.
    T: float
        The sampling rate (default = 1)

    Returns
    -------
    dict
        `tau_min`: the minimum tau value
        `rho_min`: the minimum rho value, the real-part of z-plane poles in soln
        `M_min`: the M-value related to a_min
        `success`: success bool,
        `iter_count`: the # of iterations to solve,
        `message`: rho message about the solution,
    """

    # validate entry
    if splane_poles.shape[0] <= 2:
        msg = (
            "This solver must have 3 or more complex-conjugate poles and "
            "at most one real-valued pole in order to formulate "
            "the objective. The current pole-count is {0}".format(
                splane_poles.shape[0]
            )
        )
        raise RuntimeError(msg)

    # step 1: identify poles with (min, max) imaginary parts
    working_spoles = {
        "inner": splane_poles[np.argmin(np.abs(np.imag(splane_poles)))],
        "outer": splane_poles[np.argmax(np.imag(splane_poles))],
    }

    # helpers
    #   brings p: (0, 1) to the linear scale M: (0, inf)
    p2M = lambda p: p / (1.0 - p)
    M2p = lambda M: M / (1.0 + M)

    #   returns the rho-value for the inner pole given a tau input
    calc_rho_inner = lambda tau: np.real(
        convert_analog_poles_to_digital_poles(tau, working_spoles["inner"], T)
    )

    # step 2: initialize, iterate until convergence of max-iter
    M_outer = 1e6
    max_iter = 1000
    curr_iter = 0
    rel_tol = 1e-3
    success = False

    while True:

        # sweep through sequence
        rho_outer = M2p(M_outer)
        tau_min = solve_for_tau_to_place_splane_pole_onto_zplane_real_axis(
            working_spoles["outer"], rho_outer
        )
        rho_inner = calc_rho_inner(tau_min)
        M_inner = p2M(rho_inner)

        # test for convergence, or setup next step
        if np.abs(M_outer / M_inner - 1.0) < rel_tol:
            msg = "converged after {0} iterations".format(curr_iter)
            success = True
            break
        elif curr_iter == max_iter:  # pragma: no cover
            msg = "Convergence not achieved. M: ({0},{1}), tau: {2}".format(
                M_outer, M_inner, tau_min
            )
            break
        else:
            M_outer = M_inner
            curr_iter += 1

    # pack for return
    solution = {
        "tau_min": tau_min,
        "rho_min": rho_outer,
        "M_min": p2M(rho_outer),
        "success": success,
        "iter_count": curr_iter,
        "message": msg,
    }

    return solution



"""
-------------------------------------------------------------------------------

Analog poly-ema inline slope filter design:

Captures the design details for this filter. Design details are available
in Mathematica notebook polyema_inline_slope.nb. Wavenumbers are
calculated numerically in Python in `calculate_wavenumber.py`.

-------------------------------------------------------------------------------
"""

import numpy as np
import scipy.misc
import scipy.special
import sys

from irides.design.analog import polyema_level as design_reference
from irides.tools import design_tools


# noinspection SpellCheckingInspection
def design_id() -> str:
    """
    Since `design` modules are passed around as variables in this framework,
    a `design_id` gives a unique name.

    Returns
    -------
    str
        The id of this design.
    """

    return "polyema-inline-slope"


def design_type() -> design_tools.FilterDesignType:
    """This is a slope filter"""

    return design_tools.FilterDesignType.SLOPE


def get_minimum_valid_filter_order(strict: bool = False) -> int:
    """Returns minimum valid filter order."""

    return design_tools.get_minimum_valid_filter_order(design_type(), strict)


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns array of valid filter orders based on strictness."""

    return design_tools.get_valid_filter_orders(design_type(), strict)


def integer_sequence_a000169(n: int) -> int:
    """
    Integer sequence that is used in the peak-gain calculation.

    See https://oeis.org/A000169.

    Parameters
    ----------
    n: int
        Sequence index.

    Returns
    -------
    int
        Sequence value.
    """

    return np.power(n, n - 1)


def integer_sequence_a063170(n: int) -> int:
    """
    Integer sequence related to inline polyema slope-filter
    wireframe timepoints.

    See https://oeis.org/A063170. One note: The scipy.special
    function of gammaincc(a,b) is normalized by gamma(a). To align
    with the published sequence, the gammaincc(..) function is
    multiplied by gamma(a).

    Parameters
    ----------
    n: int
        Sequence index.

    Returns
    -------
    int
        Sequence value.
    """

    return int(
        np.round(
            scipy.special.gamma(n)
            * scipy.special.gammaincc(n, n - 1)
            * np.exp(n - 1)
        )
    )


# noinspection SpellCheckingInspection
def impulse_response_t0_value(
    tau: float, filter_order: int, dt: float
) -> float:
    """
    Returns the h_m(t=0) value for inline slope filters of this type.
    For order = 1, there is, technically, a leading Dirac delta function.
    While this module falls within the continuous-time context, here the
    time discretization is accounted for.

    Parameters
    ----------
    dt: float
        Temporal discretization
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        Theoretic value of h_m(t=0).
    """

    if filter_order == 1:
        val = 1.0 / (dt * tau) - 1.0 / np.power(tau, 2)
    elif filter_order == 2:
        val = 4.0 / np.power(tau, 2)
    else:
        val = 0.0

    return val


def zero_crossing_time(tau: float, filter_order: int) -> float:
    r"""
    The zero-crossing time `tx` is when h_slope(tx) = 0. This is
    the same time when h_ref(tx) reaches its maximum.

      h_slope(t)
          ^       .
          |    ... ...
             ..       .
          ---|----|----0----|----------------------------> t
             0          .       ....
                         ... ...
                       ^    .
                       |
                      tx

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        Time tx where h_slope(tx) = 0.
    """

    return (filter_order - 1.0) / filter_order * tau


# noinspection SpellCheckingInspection,PyUnusedLocal
def moment_value_generator(moment: int, tau: float = 1.0, dt: float = 0.0):
    """
    Returns an anonymous function with the single argument (order: int),
    where order is a `filter_order`. The returned function captures the `tau`
    and `dt` arguments in this calling function.

    Note that the generated functions are not purely continuous time.
    Instead, they bridge continuous and discrete time by applying order-dt
    corrections to the moments. This is required because the CT series are
    in fact implemented in DT.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # generator defs with `tau` and `dt` capture
    # noinspection PyUnusedLocal
    def m0_gen(order: int) -> float:
        if order == 1:
            return 0.0 - dt / (2.0 * np.power(tau, 2))
        elif order == 2:
            return 0.0 + 2.0 * dt / np.power(tau, 2)
        else:
            return 0.0

    # noinspection PyUnusedLocal
    def m1_gen(order: int) -> float:
        return -1.0

    # noinspection PyUnusedLocal
    def m2_gen(order: int) -> float:
        return -2.0 * tau

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection
def autocorrelation_peak_and_stride_values(
    tau: float, filter_order: int, dt=0.0
) -> dict:
    """
    The zero-lag system autocorrelation value, k_hslope(0), is analytic
    and can be expressed in terms of its reference-filter parent. We have

                                 1
        k_hslope(0) = ----------------------- k_href(0)
                       (2 m - 3) (tau / m)^2

    where m = filter_order.

    The stride xi_5pct where sacf(lag=xi_5pct) = 0.05 is reported numerically
    (based on results from Mathematica).

    If `dt` is not None, then the discretization correction is applied to kh0.

    k_hslope(0) scales as 1 / tau^3.
    xi_5pct scales as tau.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        Order of the filter.
    dt: float
        [Optional] Temporal increment dt in h(t).

    Returns
    -------
    dict
        Peak value and 5% pct stride, appropriately scaled by tau
    """

    # compute kh0 value
    #   note: returns {'kh0': <val>, 'xi_5pct': <val>}
    ref_sacf_values = design_reference.autocorrelation_peak_and_stride_values(
        tau, filter_order
    )
    slope_sacf_values = ref_sacf_values

    # noinspection SpellCheckingInspection
    def coef_gen(order: int) -> float:
        """Analytic coefficient to reference-filter kh0"""

        return 1.0 / ((2.0 * order - 3.0) * np.power(tau / order, 2))

    # adjust kh0 value
    if filter_order > 1:
        slope_sacf_values["kh0"] *= coef_gen(filter_order)
    else:
        slope_sacf_values["kh0"] = np.nan

    # adjust for discretization
    slope_sacf_values[
        "kh0"
    ] += autocorrelation_peak_value_discretization_correction(
        tau, filter_order, dt
    )

    # set xi_5pct value, see Mathematica polyema_inline_slope.nb
    unit_sacf_xi_5pct_values = {
        1: np.nan,
        2: 2.06997,
        3: 2.18896,
        4: 2.03310,
        5: 1.87929,
        6: 1.74845,
        7: 1.63877,
        8: 1.54614,
    }
    slope_sacf_values["xi_5pct"] = unit_sacf_xi_5pct_values[filter_order] * tau

    # return
    return slope_sacf_values


# noinspection SpellCheckingInspection
def autocorrelation_peak_value_discretization_correction(
    tau: float, filter_order: int, dt: float
) -> float:
    """
    Although this is a continuous-time context, the discretization of the
    system autocorrelation function requires a `dt`-order correction. This
    correction can be computed directly by taking the Maclaurin expansion
    of the summed correlogram in terms of dt.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        Order of the filter.
    dt: float
        Time-discretization step.

    Returns
    -------
    float
        Discretization correction.
    """

    if filter_order == 1:
        return np.nan
    elif filter_order == 2:
        return 8.0 * dt / np.power(tau, 4)
    else:
        return 0.0


# noinspection SpellCheckingInspection,PyPep8Naming
def scale_correlation_stride_and_residual_values(
    tau: float, filter_order: int
) -> dict:
    """Scale decorrelation strides and residual correlations.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Scale decorrelation (scales as `tau`) and residual correlation
        (independent of `tau`).
    """

    # values from Mathematica
    unit_sscf_values = {
        2: {"sti_5pct": 6.39659, "residual": +0.05000},
        3: {"sti_5pct": 5.89306, "residual": -0.03034},
        4: {"sti_5pct": 5.47517, "residual": -0.05000},
        5: {"sti_5pct": 5.18021, "residual": -0.05000},
        6: {"sti_5pct": 4.72284, "residual": -0.05000},
        7: {"sti_5pct": 4.33074, "residual": -0.05000},
        8: {"sti_5pct": 4.01296, "residual": -0.05000},
    }

    # fetch values for order and scale
    sscf_values = unit_sscf_values[filter_order]
    sscf_values["sti_5pct"] *= tau

    # return
    return sscf_values


# noinspection PyUnusedLocal
def gain_at_dc(tau: float) -> float:
    """Filter gain at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return 0.0


# noinspection PyUnusedLocal
def phase_at_dc(tau: float) -> float:
    """Filter phase at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return np.pi / 2


def group_delay_at_dc(tau: float) -> float:
    """Filter group delay at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return design_reference.group_delay_at_dc(tau)


# noinspection SpellCheckingInspection
def frequency_of_peak_gain(tau: float, filter_order: int) -> float:
    """Returns the radial frequency at which the gain peaks."""

    return generic_peak_frequency_value("frequency", tau, filter_order)


# noinspection SpellCheckingInspection
def gain_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns gain(w_peak), the peak gain."""

    return generic_peak_frequency_value("gain", tau, filter_order)


# noinspection SpellCheckingInspection
def phase_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns angle H_delta(s) at the peak frequency."""

    return generic_peak_frequency_value("phase", tau, filter_order)


# noinspection SpellCheckingInspection
def group_delay_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns the group delay at the peak frequency."""

    return generic_peak_frequency_value("group-delay", tau, filter_order)


# noinspection SpellCheckingInspection
def generic_peak_frequency_value(
    feature: str, tau: float, filter_order: int
) -> float:
    """Returns the value of a peak-frequency feature.

    Parameters
    ----------
    feature: str
        The caller: {frequency|gain|phase|group-delay}
    tau: float
        Temporal scale of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    float
        The value associated with the caller.
    """

    fcxns = {
        "frequency": lambda _tau, order: order / np.sqrt(order - 1) / _tau,
        "gain": lambda _tau, order: np.sqrt(
            order * np.power((order - 1.0) / order, order - 1)
        )
        / _tau,
        "phase": lambda _tau, order: np.pi / 2.0
        - order * np.arctan(1.0 / np.sqrt(order - 1)),
        "group-delay": lambda _tau, order: (order - 1.0) / order * tau,
    }

    fcxn = fcxns[feature]
    value = fcxn(tau, filter_order) if filter_order > 1 else np.nan

    return value


# noinspection SpellCheckingInspection,PyPep8Naming
def wireframe(tau: float, filter_order: int):
    """Wireframe-timepoint-pair for this slope filter.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    return design_tools.generate_slope_wireframe_from_spectra(
        sys.modules[__name__], tau, filter_order
    )


# noinspection SpellCheckingInspection
def wavenumber(filter_order: int) -> float:
    """Report the wavenumber for this filter as a function of order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The wavenumber.
    """

    wavenumbers = {
        2: 0.56986212,
        3: 0.70527775,
        4: 0.73673970,
        5: 0.74872477,
        6: 0.75584547,
        7: 0.76189466,
        8: 0.76761364,
    }

    return wavenumbers[filter_order]


# noinspection SpellCheckingInspection
def uncertainty_product(filter_order: int) -> dict:
    """Dict of TW, BW, and UCP for this filter order.

    Parameters
    ----------
    filter_order: int
        The order of the filter

    Returns
    -------
    dict
        TW is temporal width, BW is spectral width, and UCP is the
        uncertainty product.
    """

    uncertainty_products = {
        2: {"TW": 0.433013, "BW": np.inf, "UCP": np.inf},
        3: {"TW": 0.5, "BW": 5.19615, "UCP": 2.59808},
        4: {"TW": 0.484123, "BW": 4.0, "UCP": 1.93649},
        5: {"TW": 0.458258, "BW": 3.87298, "UCP": 1.77482},
        6: {"TW": 0.433013, "BW": 3.92792, "UCP": 1.70084},
        7: {"TW": 0.410326, "BW": 4.04145, "UCP": 1.65831},
        8: {"TW": 0.390312, "BW": 4.17786, "UCP": 1.63067},
    }

    return uncertainty_products[filter_order]



"""
-------------------------------------------------------------------------------

Digital poly-ema inline curvature-filter signature.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.filter_signatures.digital import polyema_level as fsig_dig_l_pema
from irides.operators.digital import second_difference as fsig_2nd_diff
from irides.resources.containers.discrete_sequence import DiscreteSequence


# noinspection SpellCheckingInspection
def generate_impulse_response(
    n_start: int, n_end: int, mu: float, filter_order: int
) -> DiscreteSequence:
    """Creates a DT impulse response for the poly-ema inline curvature filter.

    Parameters
    ----------
    n_start: int
        Start sample
    n_end: int
        End sample (exclusive)
    mu: float
        First moment of the impulse response
    filter_order: int
        Filter order

    Returns
    -------
    DiscreteSequence
        Contains n-axis and v-axis
    """

    # create the level-filter and 2nd-diff components
    ds = fsig_dig_l_pema.generate_impulse_response(
        n_start, n_end, mu, filter_order
    )
    ds_2nd_diff = fsig_2nd_diff.generate_impulse_response()

    # convolve
    cand = np.convolve(ds_2nd_diff.v_axis, ds.v_axis)

    # assign result back to original, clipping last value
    ds.v_axis = cand[: ds.n_axis.shape[0]]

    # return
    return ds



"""
-------------------------------------------------------------------------------

Discrete-time test signals.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.resources.core_enumerations import TestSignalType
from irides.resources.containers.discrete_sequence import DiscreteSequence


# noinspection SpellCheckingInspection
def generate_test_signal(
    n_start: int, n_end: int, signal_type: TestSignalType,
) -> DiscreteSequence:
    """Discrete-time test signals on an integer axis.

    Parameters
    ----------
    n_start: int
        Start index.
    n_end: int
        End index.
    signal_type: TestSignalType
        Type of test signal to construct.

    Returns
    -------
    DiscreteSequence
        Contains n-axis and v-axis
    """

    # create a template
    ds = DiscreteSequence(n_start, n_end)

    # quick return if type is unknown
    if signal_type == TestSignalType.UNKNOWN:
        return ds

    # build lambda dict
    sequence_functions = {
        TestSignalType.IMPULSE: lambda n: 1.0 if n == 0 else 0.0,
        TestSignalType.STEP: lambda n: 1.0,
        TestSignalType.RAMP: lambda n: n + 1.0,
        TestSignalType.PARABOLA: lambda n: 0.5 * (n + 1.0) * (n + 2.0),
        TestSignalType.ALTERNATING: lambda n: np.power(-1.0, n),
        TestSignalType.WHITE_NOISE: lambda n: np.random.normal(0, 1, 1)[0],
    }

    # apply
    ds.v_axis[ds.i_n_zero :] = np.array(
        list(map(sequence_functions[signal_type], ds.n_axis[ds.i_n_zero :]))
    )

    # force causality

    if ds.i_n_zero is not None:
        ds.v_axis[: ds.i_n_zero] = 0.0

    # return
    return ds



"""
-------------------------------------------------------------------------------

Digital multistage-box level-filter design.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.analog import mbox_level as dsgn_alg_l_mbox
from irides.tools import design_tools


def design_id() -> str:
    """Design id of this filter."""

    return "digital-{0}".format(dsgn_alg_l_mbox.design_id())


def design_type() -> design_tools.FilterDesignType:
    """This is a level filter"""

    return dsgn_alg_l_mbox.design_type()


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns valid filter orders for this filter design."""

    return dsgn_alg_l_mbox.get_valid_filter_orders(strict)


# noinspection PyPep8Naming
def convert_mu_to_N(mu: float, filter_order: int) -> float:
    """Converts mu value to Mbox length N.

    No adjustments are made to ensure that N / m is an integer. The first
    moment of a mbox filter is
        mu = (N - m) / 2,
    thus
        N = 2 mu + m.
    """

    return 2 * mu + filter_order


def adjust_mu_to_ensure_integer_length_box(
    mu: float, filter_order: int
) -> float:
    """Adjusts `mu` to that box-length `N/m` is an integer.

    At issue is that `mu` is given, but mu = (N - m) / 2, so the
    resulting `N` may not be an integer. The strategy is to convert
    N -> N_int and then return mu* = (N_int - m) / 2.
    """

    N_cand = 2.0 * mu + filter_order
    N_int = filter_order * np.round(N_cand / filter_order).astype(int)

    return (N_int - filter_order) / 2


# noinspection PyPep8Naming
def convert_adjusted_mu_to_N_stage(mu: float, filter_order: int) -> int:
    """Returns an integral N_stage value after, in effect, adjusting `mu`."""

    N_cand = 2.0 * mu + filter_order

    return np.round(N_cand / filter_order).astype(int)


def minimum_mu_value(filter_order: int) -> float:
    """Returns the minimum mu value for a given filter order"""

    return filter_order / 2


# noinspection PyPep8Naming
def gain_adjustment(mu: float, filter_order: int) -> float:
    """Returns the gain adjustment so that the filter's level gain is unity."""

    # compute N
    N = convert_mu_to_N(mu, filter_order)

    # compute and return
    return np.power(filter_order / N, filter_order)


def initial_hn_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns analytically calculated values of h[0], h[1], and h[2]."""

    # compute N, fetch gain adj
    gain_adj = gain_adjustment(mu, filter_order)

    # compute
    h0 = gain_adj
    h1 = filter_order * gain_adj
    h2 = 0.5 * (filter_order + 1) * filter_order * gain_adj

    # pack and return
    return np.array([h0, h1, h2])


# noinspection PyPep8Naming
def moment_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns analytically calculated moment values (M0, M1, M2)"""

    # compute integer N
    N = convert_mu_to_N(mu, filter_order)
    m = filter_order

    M0 = 1.0
    M1 = 0.5 * (N - m)
    M2 = ((3 * m + 1) * N - (3 * m - 1) * m) / (6 * m) * M1

    return np.array([M0, M1, M2])


# noinspection PyPep8Naming
def full_width_value(mu: float, filter_order: int) -> np.ndarray:
    """Returns the analytically calculated FW value"""

    Mk = moment_values(mu, filter_order)

    return 2.0 * np.sqrt(Mk[2] - np.power(Mk[1], 2))


def analog_tau_value(mu: float, filter_order: int) -> float:
    """Returns `tau` associated with (`mu`, `filter_order`)"""

    # for FIR filters like mbox, tau = mu
    return mu



"""
-------------------------------------------------------------------------------

Digital poly-ema level-filter signature.

-------------------------------------------------------------------------------
"""

import numpy as np
import scipy.special

from irides.design.digital import polyema_level as dsgn_d_l_pema

from irides.tools import digital_design_tools as dd_tools
from irides.resources.containers.discrete_sequence import DiscreteSequence
from irides.tools.digital_design_tools import FrequencyBand


# noinspection SpellCheckingInspection
def generate_impulse_response(
    n_start: int,
    n_end: int,
    mu: float,
    filter_order: int,
    frequency_band: FrequencyBand = FrequencyBand.LOW,
) -> DiscreteSequence:
    """Creates a discrete-time impulse response for the poly-ema filter.

    Parameters
    ----------
    n_start: int
        Start sample
    n_end: int
        End sample (exclusive)
    mu: float
        First moment of the impulse response
    filter_order: int
        Filter order
    frequency_band: FrequencyBand
        Specifies LOW or HIGH frequency (default = LOW)

    Returns
    -------
    DiscreteSequence
        Contains n-axis and h-axis
    """

    # compute pole value
    p = dsgn_d_l_pema.convert_mu_to_pole(mu, filter_order)

    # establish the location of `p` based on the requested frequency band
    # fmt: off
    p_signed = dd_tools.\
        convert_frequency_band(np.array([p]), frequency_band)[0]
    # fmt: on

    # setup a sequence container
    ds = DiscreteSequence(n_start, n_end)

    # pull n-axis and setup hn
    i_n_zero = ds.i_n_zero
    n_0plus = ds.n_axis[i_n_zero:]

    hn = ds.v_axis

    # analytically compute response
    gain_adj = dsgn_d_l_pema.gain_adjustment(mu, filter_order)
    hn[i_n_zero:] = (
        gain_adj
        * scipy.special.binom(n_0plus + filter_order - 1, filter_order - 1)
        * np.power(p_signed, n_0plus)
    )

    # return
    return ds



"""
-------------------------------------------------------------------------------

Analog polyema inline curvature filter design:

Captures the design details for this filter. Design details are available
in Mathematica notebook polyema_inline_curvature.nb. Wavenumbers are
calculated numerically in Python in `calculate_wavenumber.py`.

-------------------------------------------------------------------------------
"""

import numpy as np
import sys

from irides.design.analog import polyema_level as design_reference
from irides.tools import design_tools


# noinspection SpellCheckingInspection
def design_id() -> str:
    """
    Since `design` modules are passed around as variables in this framework,
    a `design_id` gives a unique name.

    Returns
    -------
    str
        The id of this design.
    """

    return "pema-inline-curvature"


def design_type() -> design_tools.FilterDesignType:
    """This is a curvature filter"""

    return design_tools.FilterDesignType.CURVE


def get_minimum_valid_filter_order(strict: bool = False) -> int:
    """Returns minimum valid filter order."""

    return design_tools.get_minimum_valid_filter_order(design_type(), strict)


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns array of valid filter orders based on strictness."""

    return design_tools.get_valid_filter_orders(design_type(), strict)


# noinspection SpellCheckingInspection,PyUnusedLocal
def impulse_response_t0_value(
    tau: float, filter_order: int, dt: float
) -> float:
    """
    Returns the h_m(t=0) value for inline slope filters of this type.

    Parameters
    ----------
    dt: float
        Temporal discretization
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        Theoretic value of h_m(t=0).
    """

    # values calculated from Mathematica
    unit_t0_values = np.array(
        [
            27.0,  # m = 3
            0.0,  # m = 4
            0.0,  # m = 5
            0.0,  # m = 6
            0.0,  # m = 7
            0.0,  # m = 8
        ]
    )

    # return
    index = filter_order - get_minimum_valid_filter_order()
    return unit_t0_values[index] / np.power(tau, 3)


def impulse_response_lobe_details(tau: float, filter_order: int) -> dict:
    """
    Curvature filters have two principal crossing times, and the two positive
    lobes may capture equal or unequal area beneath.

      h_curv(t)
        ^     .                  ..
        |   .. ..             ...  ....
           .     .           .         .....
        ---|--|---0---------0-----------------------> t
           0       .       .
                    .     .
                     .. ..
                  ^    .    ^
                  |         |
                 tx-1      tx-2

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    dict
    """

    lobe_details = {
        3: {"tx1": 0.195262, "tx2": 1.13807, "imbalance": 2.90281},
        4: {"tx1": 0.316987, "tx2": 1.18301, "imbalance": 2.29374},
        5: {"tx1": 0.400000, "tx2": 1.20000, "imbalance": 2.02215},
        6: {"tx1": 0.460655, "tx2": 1.20601, "imbalance": 1.86347},
        7: {"tx1": 0.507216, "tx2": 1.20707, "imbalance": 1.75752},
        8: {"tx1": 0.544281, "tx2": 1.20572, "imbalance": 1.68087},
    }

    ans_dict = lobe_details[filter_order]
    ans_dict["tx1"] *= tau
    ans_dict["tx2"] *= tau

    return ans_dict


# noinspection SpellCheckingInspection,PyUnusedLocal
def moment_value_generator(moment: int, tau: float = 1.0, dt: float = 0.0):
    """
    Returns an anonymous function with the single argument (order: int),
    where order is a `filter_order`. The returned function captures the `tau`
    and `dt` arguments in this calling function.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    tau: float
        Temporal scale of the filter (unused)
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # generator defs with `tau` and `dt` capture
    # noinspection PyUnusedLocal
    def m0_gen(order: int) -> float:
        if order == 3:
            return 0.0 + (27.0 / 2.0) * dt / np.power(tau, 3)
        else:
            return 0.0

    # noinspection PyUnusedLocal
    def m1_gen(order: int) -> float:
        return 0.0

    # noinspection PyUnusedLocal
    def m2_gen(order: int) -> float:
        return 2.0

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection, PyUnusedLocal
def autocorrelation_peak_and_stride_values(
    tau: float, filter_order: int, dt=0.0
) -> dict:
    """
    System autocorrelation peak and stride values. Peak value is kh(0),
    and the stride comes from the trailing kh(xi_5pct) / kh(0) = 5%.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter (note: user is required to validate that
        the order is supported).
    dt: float
        [Optional] Temporal increment dt in h(t).

    Returns
    -------
    dict
        Peak value (scaled by 1/tau^5) and 5% pct stride (scaled by tau)
    """

    # values from Mathematica
    unit_sacf_values = {
        3: {"kh0": 45.5625, "xi_5pct": 1.85385, "residual_acf": 0.78433},
        4: {"kh0": 32.0000, "xi_5pct": 1.69782, "residual_acf": 2.24215},
        5: {"kh0": 36.6211, "xi_5pct": 1.56878, "residual_acf": 3.41986},
        6: {"kh0": 45.5625, "xi_5pct": 1.46340, "residual_acf": 4.35727},
        7: {"kh0": 57.4458, "xi_5pct": 1.37610, "residual_acf": 5.11265},
        8: {"kh0": 72.0000, "xi_5pct": 1.30253, "residual_acf": 5.73130},
    }

    # fetch values for order and scale
    sacf_values = unit_sacf_values[filter_order]
    # adjust for temporal discretization
    if filter_order == 3:
        sacf_values["kh0"] += 729.0 / 2.0 * dt / tau
    sacf_values["kh0"] /= np.power(tau, 5)
    sacf_values["xi_5pct"] *= tau

    # return
    return sacf_values


# noinspection SpellCheckingInspection,PyPep8Naming
def scale_correlation_stride_and_residual_values(
    tau: float, filter_order: int
) -> dict:
    """Scale decorrelation strides and residual correlations.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Scale decorrelation (scales as `tau`) and residual correlation
        (independent of `tau`).
    """

    # values from Mathematica
    unit_sscf_values = {
        3: {"sti_5pct": 3.07384, "residual": +0.05000},
        4: {"sti_5pct": 3.54070, "residual": -0.01764},
        5: {"sti_5pct": 2.69320, "residual": -0.04482},
        6: {"sti_5pct": 2.95804, "residual": -0.05000},
        7: {"sti_5pct": 2.89426, "residual": -0.05000},
        8: {"sti_5pct": 2.78014, "residual": -0.05000},
    }

    # fetch values for order and scale
    sscf_values = unit_sscf_values[filter_order]
    sscf_values["sti_5pct"] *= tau

    # return
    return sscf_values


# noinspection PyUnusedLocal
def gain_at_dc(tau: float) -> float:
    """Filter gain at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return 0.0


# noinspection PyUnusedLocal
def phase_at_dc(tau: float) -> float:
    """Filter phase at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return np.pi


def group_delay_at_dc(tau: float) -> float:
    """Filter group delay at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return design_reference.group_delay_at_dc(tau)


# noinspection SpellCheckingInspection
def frequency_of_peak_gain(tau: float, filter_order: int) -> float:
    """Returns the radial frequency at which the gain peaks."""

    return generic_peak_frequency_value("frequency", tau, filter_order)


# noinspection SpellCheckingInspection
def gain_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns gain(w_peak), the peak gain."""

    return generic_peak_frequency_value("gain", tau, filter_order)


# noinspection SpellCheckingInspection,PyUnusedLocal
def phase_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns angle H_delta(s) at the peak frequency."""

    return generic_peak_frequency_value("phase", tau, filter_order)


# noinspection SpellCheckingInspection
def group_delay_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns the group delay at the peak frequency."""

    return generic_peak_frequency_value("group-delay", tau, filter_order)


# noinspection SpellCheckingInspection
def generic_peak_frequency_value(
    feature: str, tau: float, filter_order: int
) -> float:
    """Returns the value of a peak-frequency feature.

    Parameters
    ----------
    feature: str
        The caller: {frequency|gain|phase|group-delay}
    tau: float
        Temporal scale of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    float
        The value associated with the caller.
    """

    fcxns = {
        "frequency": lambda _tau, order: order
        / np.sqrt((order - 2.0) / 2.0)
        / _tau,
        "gain": lambda _tau, order: 2.0
        * np.sqrt(
            np.power(order, 2) * np.power((order - 2.0) / order, order - 2)
        )
        / np.power(_tau, 2),
        "phase": lambda _tau, order: np.pi
        - order * np.arctan(1.0 / np.sqrt((order - 2.0) / 2.0)),
        "group-delay": lambda _tau, order: (order - 2.0) / order * tau,
    }

    fcxn = fcxns[feature]
    value = fcxn(tau, filter_order) if filter_order > 1 else np.nan

    return value


# noinspection SpellCheckingInspection,PyPep8Naming
def wireframe(tau: float, filter_order: int):
    """Wireframe-timepoint-pair for this slope filter.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    return design_tools.generate_curvature_wireframe_from_spectra(
        sys.modules[__name__], tau, filter_order
    )


# noinspection SpellCheckingInspection
def wavenumber(filter_order: int) -> float:
    """Report the wavenumber for this filter as a function of order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The wavenumber.
    """

    wavenumbers = {
        3: 0.50644018,
        4: 0.65761864,
        5: 0.70048551,
        6: 0.71784811,
        7: 0.72791012,
        8: 0.73608002,
    }

    return wavenumbers[filter_order]


# noinspection SpellCheckingInspection
def uncertainty_product(filter_order: int) -> dict:
    """Dict of TW, BW, and UCP for this filter order.

    Parameters
    ----------
    filter_order: int
        The order of the filter

    Returns
    -------
    dict
        TW is temporal width, BW is spectral width, and UCP is the
        uncertainty product.
    """

    uncertainty_products = {
        3: {"TW": 0.254588, "BW": np.inf, "UCP": np.inf},
        4: {"TW": 0.330719, "BW": 8.94427, "UCP": 2.95804},
        5: {"TW": 0.341565, "BW": 6.45497, "UCP": 2.20479},
        6: {"TW": 0.336788, "BW": 6.0, "UCP": 2.02073},
        7: {"TW": 0.327327, "BW": 5.91608, "UCP": 1.93649},
        8: {"TW": 0.316639, "BW": 5.96285, "UCP": 1.88807},
    }

    return uncertainty_products[filter_order]



"""
-------------------------------------------------------------------------------

Tools to build transfer functions and spectra.

-------------------------------------------------------------------------------
"""

from typing import Callable
import numpy as np


# noinspection SpellCheckingInspection,PyPep8Naming
def truncate_and_normalize_sacf_correlogram(
    sacf: np.ndarray, kh0: float, normalize: bool, xi_start: float
) -> np.ndarray:
    """Truncates sacf to xi >= xi_start, normalizes by kh0 if requested.

    Parameters
    ----------
    sacf: np.ndarray
        The system autocorrelation correlogram [n, 2] format.
    kh0: float
        The sacf(0) value.
    normalize: bool
        The normalization flag.
    xi_start: float
        Starting point along the xi axis

    Returns
    -------
    np.ndarray
        Modified correlogram.
    """

    # infer
    dxi = np.min(np.diff(sacf[:, 0]))

    # truncate
    sacf = sacf[xi_start - dxi / 2.0 <= sacf[:, 0], :]

    # normalize is requested
    if normalize:
        sacf[:, 1] /= kh0

    # return
    return sacf


# noinspection SpellCheckingInspection,PyPep8Naming
def apply_leading_jw_coefficient_to_transfer_function(
    H: np.ndarray, power: int = 1
) -> np.ndarray:
    """Multiplies the transfer function by a power of jw.

    Parameters
    ----------
    H: np.ndarray
        Current transfer function, [n, 2] layout
    power: int
        Applies (jw)^power to the transfer function. Default = 1.

    Returns
    -------
    np.ndarray
        Updated transfer function
    """

    f_axis = H[:, 0]
    coef = (1j * 2.0 * np.pi) * f_axis
    H[:, 1] *= np.power(coef, power)

    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def apply_leading_jw_coefficient_to_phase_spectrum(
    H: np.ndarray, power: int = 1
) -> np.ndarray:
    """Adds multiple of np.pi / 2.0 to the phase.

    Parameters
    ----------
    H: np.ndarray
        Current phase spectrum, [n, 2] layout
    power: int
        Applies (np.pi/2) * power to the phase. Default = 1.

    Returns
    -------
    np.ndarray
        Updated phase spectrum
    """

    phase_shift = (np.pi / 2.0) * power
    H[:, 1] += phase_shift

    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def make_callable_transfer_function_with_leading_jw_power_term(
    f_sig_ref, filter_order: int, power: int
) -> Callable[[float, float], complex]:
    """Creates new callable with same signature with leading (jw)^power.

    Parameters
    ----------
    f_sig_ref
        Name of reference filter-signature module.
    filter_order: int
        The filter order, which is captured by the definition herein.
    power: int
        Power to raise (jw) term

    Returns
    -------
    Callable[[float, float], complex]
    """
    H_ref = f_sig_ref.make_callable_transfer_function(filter_order)

    # noinspection PyPep8Naming
    def H_amended(f: float, tau: float = 1.0) -> complex:

        return np.power(1j * 2.0 * np.pi * f, power) * H_ref(f, tau)

    return H_amended



"""
-------------------------------------------------------------------------------

Analog polynomial ema:

Implements the temporal, spectral, autocorrelation and pole / zero features
for this continuous-time level filter.

-------------------------------------------------------------------------------
"""

from typing import Callable
import numpy as np
import scipy.special
import sys

from irides.design.analog import polyema_level as design
from irides.tools import impulse_response_builder_tools
from irides.tools import impulse_response_tools
from irides.tools import transfer_function_tools


# noinspection SpellCheckingInspection
def generate_impulse_response(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Creates a continuous-time-like impulse response for the poly-ema
    level filter.

    The impulse response of an mth order filter reads

                        1          /    t     \ (m-1)
        h(t) = ------------------ | ---------- |      e^(-t / (tau / m)) u(t),
               (tau / m) Gamma(m)  \  tau / m /

    where tau is the first-moment of the h(t).

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order .

    Returns
    -------
    np.ndarray
        A time series array (n, 2) having columns [t_axis, h_pema].
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # make time series template
    # fmt: off
    time_series, i_t_0plus = impulse_response_builder_tools. \
        make_impulse_response_template(
            t_start, t_end, dt
        )
    # fmt: on

    # alias
    t_axis = time_series[:, 0]
    t_0plus = t_axis[i_t_0plus:]
    h = time_series[:, 1]

    # compute impulse response t >= 0
    m = filter_order
    tau_m = design.tau_per_stage(tau, m)

    h[i_t_0plus:] = (
        (1.0 / (tau_m * scipy.special.gamma(m)))
        * np.power(t_0plus / tau_m, m - 1)
        * np.exp(-t_0plus / tau_m)
    )

    # return
    return time_series


# noinspection SpellCheckingInspection
def generate_sacf_correlogram(
    xi_start: float,
    xi_end: float,
    dxi: float,
    tau: float,
    filter_order: int,
    normalize=False,
) -> np.ndarray:
    """
    Computes the system autocorrelation response from the correlation
    of the impulse response with itself.

    Parameters
    ----------
    xi_start: float
        Start lag, |xi_start| <= |xi_end|
    xi_end: float
        End lag
    dxi: float
        Lag increment
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.
    normalize: bool
        T/F whether to normalize the sacf spectrum so sacf(0) = 1.

    Returns
    -------
    np.ndarray
        System autocorrelation (n, 2) np.array w/ cols [xi, sacf-value].
    """

    # fetch impulse response
    ir = generate_impulse_response(0.0, xi_end, dxi, tau, filter_order)

    # calc sacf
    sacf = impulse_response_tools.calculate_auto_correlation_function(ir)

    # truncate at xi_start
    sacf = sacf[xi_start - dxi / 2.0 <= sacf[:, 0], :]

    # normalize by theoretic peak value if requested
    if normalize:
        kh0 = design.autocorrelation_peak_and_stride_values(tau, filter_order)[
            "kh0"
        ]
        sacf[:, 1] /= kh0

    return sacf


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_transfer_function(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Computes the transfer function H_pema(f) of the polynomial ema.
    The defining transfer function is

                             1
        H(tau s) = --------------------- .
                    (1 + (tau / m) s)^m

    In general, s = omega + j w, a complex value. Here, however, omega = 0,
    so /s/ is replaced by /j w/. Note that w = 2 pi f.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        (n, 2), dtype=complex
        The transfer function with cols [f_axis, H_pema(f)].
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # alias
    m = filter_order

    # make transfer-function panel
    npts = np.arange(f_start, f_end, df).shape[0]
    H_pema = np.ndarray([npts, 2], dtype=complex)

    # setup f_axis
    H_pema[:, 0] = np.arange(f_start, f_end, df)
    f_axis = H_pema[:, 0]

    # make f_axis and sTm vector
    jwtaum = (1j * 2.0 * np.pi * tau / m) * f_axis

    # calculate transfer function
    H_pema[:, 1] = 1.0 / np.power(1.0 + jwtaum, m)

    # return
    return H_pema


# noinspection SpellCheckingInspection,PyPep8Naming
def make_callable_transfer_function(
    filter_order: int,
) -> Callable[[float, float], complex]:
    """Returns a callable function with signature (f, tau) -> complex.

    This transfer function is restricted to the jw axis:

        H(s) -> H(j w).

    Parameters
    ----------
    filter_order: int
        The filter order, which is captured by the definition herein.

    Returns
    -------
    Callable[[float, float], complex]
        A callable function (f: float, tau: float) -> complex
    """

    # noinspection PyPep8Naming
    def H(f: float, tau: float = 1.0) -> complex:
        """Compute complex xfer fcxn H at freq `f` with optional scale tau.

        Note that `filter_order` is captured, leaving only (f, tau) as free
        variables.

        Parameters
        ----------
        f: float
            Start cyclic frequency.
        tau: float
            Temporal scaling, which ='s the first moment of the final filter.

        Returns
        -------
        complex
            H(f, tau) value.
        """

        jw = 1j * 2.0 * np.pi * f

        return np.power(1.0 + jw * tau / filter_order, -filter_order)

    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_sscf_correlogram(
    sti_start: float, sti_end: float, dsti: float, filter_order: int
) -> np.ndarray:
    """
    Computes the normalized system scale-correlation correlogram from the
    filter's transfer function.

    No sti-normalization correction is necessary for level filters.

    Parameters
    ----------
    sti_start: float
        Start scale, |sti_start| <= |sti_end|
    sti_end: float
        End scale
    dsti: float
        Scale increment
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        System scale-correlation normalized (n, 2)
        np.array w/ cols [sti, sscf-value].
    """

    # correction function
    sti_norm_correction_fcxn = lambda sti: 1.0

    # return sscf-norm np-array
    return transfer_function_tools.calculate_scale_correlation_function(
        sti_start,
        sti_end,
        dsti,
        filter_order,
        sys.modules[__name__],
        sti_norm_correction_fcxn,
    )


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_phase_spectrum(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Computes the phase of H_pema(f) from the analytic form

        angle_H_pema(f) = -m arctan( j 2 pi f tau / m ).

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order (number of stages).  (unused)

    Returns
    -------
    np.ndarray
        (n, 2), dtype=complex
        The phase spectrum with cols [f_axis, angle_H_pema(f)].
    """

    # make phase-spectrum panel
    npts = np.arange(f_start, f_end, df).shape[0]
    angle_H_pema = np.ndarray([npts, 2], dtype=float)

    # setup f_axis
    angle_H_pema[:, 0] = np.arange(f_start, f_end, df)
    f_axis = angle_H_pema[:, 0]

    # compute the phase
    m = filter_order
    angle_H_pema[:, 1] = -m * np.arctan(2.0 * np.pi * (tau / m) * f_axis)

    return angle_H_pema


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_spectra(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    """Generate gain, phase and group-delay spectra.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        Spectra panel [n_pts, 4] with cols [f, g(f), phase(f), grpdly(f)].
    """

    # fetch the transfer function for gain calculation
    H: np.ndarray = generate_transfer_function(
        f_start, f_end, df, tau, filter_order
    )

    # fetch the phase spectrum for group-delay calculation
    phase_H: np.ndarray = generate_phase_spectrum(
        f_start, f_end, df, tau, filter_order
    )

    # return spectra
    return transfer_function_tools.calculate_spectra(H, phase_H)


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_wireframe(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    """Returns a wireframe equivalent for this filter.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    WireframeContinuousTime
        A wireframe container.
    """

    # fetch the wireframe, create a timeseries template
    wf = design.wireframe(tau, filter_order)
    ts = impulse_response_builder_tools.discretize_wireframe_onto_time_grid(
        t_start, t_end, dt, wf
    )

    return ts


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def generate_poles_and_zeros(
    tau: float, filter_order: int, f_start=0.0, f_end=0.0
) -> dict:
    """
    Calculates poles and zeros for the polyema filter in the s-plane.
    There are no zeros, and the only pole(s) is an mth-order degenerate
    pole at

        s = sigma + jw = -(tau / m)^-1 .

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.
    f_start: float
        Start cyclic frequency (not used in this signature).
    f_end: float
        End cyclic frequency (not used in this signature).

    Returns
    -------
    dict
        Dictionary with keys ('poles', 'zeros') and values as np.ndarray
        with dtype=complex.
    """

    pole_location = -1.0 / (tau / filter_order) + 1j * 0.0

    # fill in poles_zeros dict
    poles_zeros = {
        "poles": np.array([pole_location]),
        "zeros": np.ndarray((0, 0), dtype=complex),
    }

    # return
    return poles_zeros



"""
-------------------------------------------------------------------------------

Tools to compute realized moments and autocorrelation functions from a
numerically calculated impulse response.

-------------------------------------------------------------------------------
"""

import numpy as np
from scipy import signal


# noinspection SpellCheckingInspection,PyPep8Naming
def calculate_impulse_response_moment(
    impulse_response: np.array, moment: int
) -> float:
    """
    Calculate a moment of the input impulse response.

        M_k = sum( n^moment * h ) * dt.

    Parameters
    ----------
    impulse_response: np.array
        Impulse response, where [:, 0] is the time axis and [:, 1] is the
        weight axis.
    moment: int
        The moment to calculate.

    Returns
    -------
    moment_value: float
        The value of the `moment` moment.
    """

    # references
    t = impulse_response[:, 0]
    h = impulse_response[:, 1]

    # infer dt, the increment of time
    dt = min(np.diff(t))

    # calculate moment
    mk = np.power(t, moment).transpose().dot(h) * dt

    # return
    return mk


# noinspection SpellCheckingInspection,PyPep8Naming
def calculate_impulse_response_full_width(impulse_response: np.array) -> float:
    """
    Calculates the full width of the input impulse response.

        FW = 2 sqrt(M2 - M1^2).

    Parameters
    ----------
    impulse_response: np.array
        Impulse response, where [:, 0] is the time axis and [:, 1] is the
        weight axis.

    Returns
    -------
    fw: float
        The full width of the impulse response.
    """

    # calculate the location and M2
    M1 = calculate_impulse_response_moment(impulse_response, 1)
    M2 = calculate_impulse_response_moment(impulse_response, 2)

    # calculate the
    arg = M2 - np.power(M1, 2)
    fw = 2.0 * np.sqrt(arg) if arg >= 0.0 else np.nan

    # return
    return fw


# noinspection SpellCheckingInspection,PyPep8Naming
def calculate_auto_correlation_function(impulse_response: np.array) -> np.array:
    """
    Calculates the system autocorrelation spectrum from the input impulse
    response.

        SACF = conv(h[-t], h[t]) * dt.

    Parameters
    ----------
    impulse_response: np.array
        Impulse response, where [:, 0] is the time axis and [:, 1] is the
        weight axis.

    Returns
    -------
    sacf: np.array
        The system autocorrelation spectrum of the input impulse response.

    """

    # references
    t = impulse_response[:, 0]
    h = impulse_response[:, 1]

    # infer dt, the increment of time
    dt: float = min(np.diff(t))

    # compute number of output points and create output panel
    npts = 2 * t.shape[0] - 1
    acf = np.ndarray([npts, 2], dtype="float")

    # compute the lags axis
    acf[:, 0] = np.array(range(npts)) * dt + (min(t) - max(t))

    # The ACF is the convolution of impulse_response(t)
    # with impulse_response(-t), scaled by dt.
    acf[:, 1] = signal.convolve(h[::-1], h, mode="full", method="direct") * dt

    # return
    return acf


def get_valid_moments() -> np.ndarray:
    """Array of valid moments in this framework.

    Returns
    -------
    np.ndarray
        Array of valid moments.
    """

    return np.array([0, 1, 2])


# noinspection SpellCheckingInspection,PyPep8Naming
def validate_filter_order_or_die(
    design, filter_order: int, strict: bool = False
):
    """
    This function either throws or returns nothing. If it returns, the
    calling code continues knowing that the filter-order is valid.

    Parameters
    ----------
    design:
        An imported design module that impl's get_valid_filter_orders().
    filter_order: int
        The filter order.
    strict: bool
        Whether to enforce strict orders.
    """

    valid_filter_orders = design.get_valid_filter_orders(strict)

    m_min = min(valid_filter_orders)
    m_max = max(valid_filter_orders)

    if not m_min <= filter_order <= m_max:
        msg = "filter order must be within [1, {0}] inclusive. order: {1}".format(
            m_max, filter_order
        )
        raise IndexError(msg)



"""
-------------------------------------------------------------------------------

Analog polyema composite-slope filter:

Implements the temporal, spectral, autocorrelation and pole/zero features
for this continuous-time composite-slope filter.

-------------------------------------------------------------------------------
"""

from typing import Callable
import numpy as np
import sys

from irides.design.analog import polyema_composite_slope as design

from irides.filter_signatures.analog import polyema_level as f_sig_ref

from irides.tools import impulse_response_builder_tools
from irides.tools import transfer_function_builder_tools
from irides.tools import impulse_response_tools
from irides.tools import transfer_function_tools


# noinspection SpellCheckingInspection
def generate_impulse_response(
    t_start: float,
    t_end: float,
    dt: float,
    tau: float,
    filter_order: int,
    arm_ratio=0.5,
) -> np.ndarray:
    """

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.
    arm_ratio: float
        Ratio between tp / tm (default = 1/2)

    Returns
    -------
    np.ndarray
        A time series array (n, 2) having columns [t_axis, h].
    """

    # convert parameters to (tau+, tau-) literals
    tp, tm = design.convert_armratio_order_params_to_tp_tm_params(
        arm_ratio, tau, filter_order
    )

    # calculate gain adjustment
    gain_adj = 1.0 / (tm - tp)

    # calculate impulse response for each arm
    ts_pos_arm = f_sig_ref.generate_impulse_response(
        t_start, t_end, dt, tp, filter_order
    )
    ts_neg_arm = f_sig_ref.generate_impulse_response(
        t_start, t_end, dt, tm, filter_order
    )

    # construct composite-slope impulse response
    time_series = ts_pos_arm
    time_series[:, 1] = gain_adj * (ts_pos_arm[:, 1] - ts_neg_arm[:, 1])

    return time_series


# noinspection SpellCheckingInspection
def generate_sacf_correlogram(
    xi_start: float,
    xi_end: float,
    dxi: float,
    tau: float,
    filter_order: int,
    arm_ratio=0.5,
    normalize=False,
) -> np.ndarray:
    """
    Computes the system autocorrelation response from the correlation
    of the impulse response with itself.

    Parameters
    ----------
    xi_start: float
        Start lag, |xi_start| <= |xi_end|
    xi_end: float
        End lag
    dxi: float
        Lag increment
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.
    arm_ratio: float
        Ratio between tp / tm (default = 1/2)
    normalize: bool
        T/F whether to normalize the sacf spectrum so sacf(0) = 1.

    Returns
    -------
    np.ndarray
        System autocorrelation (n, 2) np.array w/ cols [xi, sacf-value].
    """

    # fetch impulse response
    ir = generate_impulse_response(
        0.0, xi_end, dxi, tau, filter_order, arm_ratio
    )

    # calc sacf
    sacf = impulse_response_tools.calculate_auto_correlation_function(ir)

    # fetch kh0
    is_default_arm_ratio = np.round(arm_ratio / 0.5, 4) == 1.0
    if is_default_arm_ratio:
        kh0 = design.autocorrelation_peak_and_stride_values(
            tau, filter_order, dxi
        )["kh0"]
    else:
        i_xi0 = np.where(0.0 - dxi / 2.0 < sacf[:, 0])[0][0]
        kh0 = sacf[i_xi0, 1]  # empirical value

    # post process
    # fmt: off
    sacf = transfer_function_builder_tools. \
        truncate_and_normalize_sacf_correlogram(
            sacf, kh0, normalize, xi_start
        )
    # fmt: on

    return sacf


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_transfer_function(
    f_start: float,
    f_end: float,
    df: float,
    tau: float,
    filter_order: int,
    arm_ratio=0.5,
) -> np.ndarray:
    r"""
    Computes the transfer function H_slope(f) based on the
    underlying level filter H_lvl(f):

        H_slope(tau s) = gain_adj (H_lvl(tau_+ s) - H_lvl(tau_- s).

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.
    arm_ratio: float
        Ratio between tp / tm (default = 1/2)

    Returns
    -------
    np.ndarray
        (n, 2), dtype=complex
        The transfer function with cols [f_axis, H_slope(f)].
    """

    # convert parameters to (tau+, tau-) literals
    tp, tm = design.convert_armratio_order_params_to_tp_tm_params(
        arm_ratio, tau, filter_order
    )

    # calculate gain adjustment
    gain_adj = 1.0 / (tm - tp)

    # fetch transfer function for each arm
    H_pos_arm = f_sig_ref.generate_transfer_function(
        f_start, f_end, df, tp, filter_order
    )
    H_neg_arm = f_sig_ref.generate_transfer_function(
        f_start, f_end, df, tm, filter_order
    )

    # construct the comp-slope xfer function
    H = H_pos_arm
    H[:, 1] = gain_adj * (H_pos_arm[:, 1] - H_neg_arm[:, 1])

    # return
    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def make_callable_transfer_function(
    filter_order: int, arm_ratio=0.5
) -> Callable[[float, float], complex]:
    """Returns a callable function with signature (f, tau) -> complex.

    This transfer function is restricted to the jw axis:

        H(s) -> H(j w).

    Parameters
    ----------
    filter_order: int
        The filter order, which is captured by the definition herein.
    arm_ratio: float
        Ratio between tp / tm (default = 1/2)

    Returns
    -------
    Callable[[float, float], complex]
        A callable function (f: float, tau: float) -> complex
    """

    # convert parameters to (tau+, tau-) literals (unit tau)
    tp, tm = design.convert_armratio_order_params_to_tp_tm_params(
        arm_ratio, 1.0, filter_order
    )

    # calc gain adjustment
    gain_adj = 1.0 / (tm - tp)

    H_ref = f_sig_ref.make_callable_transfer_function(filter_order)

    # noinspection PyPep8Naming
    def H(f: float, tau: float = 1.0) -> complex:
        """Compute complex xfer fcxn H at freq `f` with optional scale tau.

        Note that `filter_order` and `arm_ratio` are captured,
        leaving only (f, tau) as free variables.

        Parameters
        ----------
        f: float
            Start cyclic frequency.
        tau: float
            Temporal scaling, which ='s the first moment of the final filter.

        Returns
        -------
        complex
            H(f, tau) value.
        """

        return gain_adj * (H_ref(f, tp * tau) - H_ref(f, tm * tau))

    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_sscf_correlogram(
    sti_start: float, sti_end: float, dsti: float, filter_order: int
) -> np.ndarray:
    """
    Computes the normalized system scale-correlation correlogram from the
    filter's transfer function.

    Note: a composite-slope filter takes the difference between two
    level filters,

        H_slope(s) = gain_adj (H_ref(tau_+ s) - H_ref(tau_- s))

    and the sscf integral covers

        int_reals H_slope(s) H_slope(-s sti) ds.

    Unlike inline-slope filters, no `sti` correction is necessary.

    Parameters
    ----------
    sti_start: float
        Start scale, |sti_start| <= |sti_end|
    sti_end: float
        End scale
    dsti: float
        Scale increment
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        System scale-correlation normalized (n, 2)
        np.array w/ cols [sti, sscf-value].
    """

    # correction function
    sti_norm_correction_fcxn = lambda sti: 1.0

    # return sscf-norm np-array
    return transfer_function_tools.calculate_scale_correlation_function(
        sti_start,
        sti_end,
        dsti,
        filter_order,
        sys.modules[__name__],
        sti_norm_correction_fcxn,
    )


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_phase_spectrum(
    f_start: float,
    f_end: float,
    df: float,
    tau: float,
    filter_order: int,
    arm_ratio=0.5,
) -> np.ndarray:
    """The phase of the transfer function.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.
    arm_ratio: float
        Ratio between tp / tm (default = 1/2)

    Returns
    -------
    np.ndarray
        The transfer function with cols [f_axis, angle_H(f)].
    """

    # convert parameters to (tau+, tau-) literals
    tp, tm = design.convert_armratio_order_params_to_tp_tm_params(
        arm_ratio, tau, filter_order
    )

    # fetch transfer function for each arm
    H_pos_arm = f_sig_ref.generate_transfer_function(
        f_start, f_end, df, tp, filter_order
    )
    H_neg_arm = f_sig_ref.generate_transfer_function(
        f_start, f_end, df, tm, filter_order
    )

    # alias
    f_axis = np.real(H_pos_arm[:, 0])
    H_pos = H_pos_arm[:, 1]
    H_neg = H_neg_arm[:, 1]

    # compute per-arm gain and phase
    g_pos = np.abs(H_pos)
    phi_pos = np.arctan2(np.imag(H_pos), np.real(H_pos))

    g_neg = np.abs(H_neg)
    phi_neg = np.arctan2(np.imag(H_neg), np.real(H_neg))

    # compute angle of composite
    im = g_pos * np.sin(phi_pos) - g_neg * np.sin(phi_neg)
    re = g_pos * np.cos(phi_pos) - g_neg * np.cos(phi_neg)
    angle = np.arctan2(im, re)

    # correct for f=0 value (given that the value comes thru a limit)
    i_f0 = np.where(0.0 - df / 2.0 < f_axis)[0][0]
    angle[i_f0] = design.phase_at_dc(tau)

    # unwrap
    angle = np.unwrap(angle)

    # force continuity across w=0 (if applicable)
    if i_f0 > 0:
        err = np.pi / 100
        if angle[i_f0] - angle[i_f0 - 1] > np.pi - err:
            angle[:i_f0] += np.pi

    # pack the return array, apply .unwrap to the angle
    angle_H = np.zeros_like(H_pos_arm, dtype=float)
    angle_H[:, 0] = np.real(H_pos_arm[:, 0]).copy()
    angle_H[:, 1] = angle

    # return
    return angle_H


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_spectra(
    f_start: float,
    f_end: float,
    df: float,
    tau: float,
    filter_order: int,
    arm_ratio=0.5,
) -> np.ndarray:
    """Generate gain, phase and group-delay spectra.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.
    arm_ratio: float
        Ratio between tp / tm (default = 1/2)

    Returns
    -------
    np.ndarray
        Spectra panel [n_pts, 4] with cols [f, g(f), phase(f), grpdly(f)].
    """

    # fetch the transfer function for gain calculation
    H: np.ndarray = generate_transfer_function(
        f_start, f_end, df, tau, filter_order, arm_ratio
    )

    # fetch the phase spectrum for group-delay calculation
    phase_spectrum = generate_phase_spectrum(
        f_start, f_end, df, tau, filter_order, arm_ratio
    )

    # return spectra
    return transfer_function_tools.calculate_spectra(H, phase_spectrum)


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_wireframe(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    """Returns a wireframe equivalent for this filter.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    WireframeContinuousTime
        A wireframe container.
    """

    # fetch the wireframe, create a timeseries template
    wf = design.wireframe(tau, filter_order)
    ts = impulse_response_builder_tools.discretize_wireframe_onto_time_grid(
        t_start, t_end, dt, wf
    )

    return ts
